# sage_setup: distribution = sagemath-gap
r"""
Rubik's cube group functions

.. _sec-rubik:

.. NOTE::

    "Rubiks cube" is trademarked. We shall omit the trademark
    symbol below for simplicity.

NOTATION:

`B` denotes a clockwise quarter turn of the back face, `D`
denotes a clockwise quarter turn of the down face, and similarly for
`F` (front), `L` (left), `R` (right), and `U` (up). Products of moves are read
right to left, so for example, `R \cdot U` means move `U` first and then `R`.

See ``CubeGroup.parse()`` for all possible input notations.

The "Singmaster notation":

- moves: `U, D, R, L, F, B` as in the
  diagram below,

- corners: `xyz` means the facet is on face `x` (in `R,F,L,U,D,B`) and the
  clockwise rotation of the corner sends `x-y-z`

- edges: `xy` means the facet is on face `x` and a flip of the edge sends
  `x-y`.

::

    sage: rubik = CubeGroup()
    sage: rubik.display2d("")
                 ┌──────────────┐
                 │  1    2    3 │
                 │  4   top   5 │
                 │  6    7    8 │
    ┌────────────┼──────────────┼─────────────┬────────────┐
    │  9  10  11 │ 17   18   19 │ 25   26  27 │ 33  34  35 │
    │ 12 left 13 │ 20  front 21 │ 28 right 29 │ 36 rear 37 │
    │ 14  15  16 │ 22   23   24 │ 30   31  32 │ 38  39  40 │
    └────────────┼──────────────┼─────────────┴────────────┘
                 │ 41   42   43 │
                 │ 44 bottom 45 │
                 │ 46   47   48 │
                 └──────────────┘

AUTHORS:

- David Joyner (2006-10-21): first version

- David Joyner (2007-05): changed faces, added legal and solve

- David Joyner(2007-06): added plotting functions

- David Joyner (2007, 2008): colors corrected, "solve" rewritten
  (again),typos fixed.

- Robert Miller (2007, 2008): editing, cleaned up display2d

- Robert Bradshaw (2007, 2008): RubiksCube object, 3d
  plotting.

- David Joyner (2007-09): rewrote docstring for CubeGroup's
  "solve".

- Robert Bradshaw (2007-09): Versatile parse function for
  all input types.

- Robert Bradshaw (2007-11): Cleanup.

REFERENCES:

- Cameron, P., Permutation Groups. New York: Cambridge
  University Press, 1999.

- Wielandt, H., Finite Permutation Groups.
  New York: Academic Press, 1964.

- Dixon, J. and Mortimer, B.,
  Permutation Groups, Springer-Verlag, Berlin/New York, 1996.

- Joyner, D., Adventures in Group Theory, Johns Hopkins Univ Press,
  2002.
"""

# *****************************************************************************
#       Copyright (C) 2006 David Joyner <wdjoyner@gmail.com>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#                  https://www.gnu.org/licenses/
# *****************************************************************************

from sage.groups.perm_gps.permgroup import PermutationGroup_generic
from random import randint

from sage.structure.sage_object import SageObject
from sage.structure.richcmp import richcmp, richcmp_method
from sage.libs.gap.libgap import libgap
from sage.rings.real_double import RDF
from sage.groups.perm_gps.permgroup_element import PermutationGroupElement
from sage.misc.lazy_import import lazy_import
lazy_import("sage.plot.polygon", "polygon")
lazy_import("sage.plot.text", "text")
pi = RDF.pi()
from sage.rings.rational_field import QQ

lazy_import("sage.plot.plot3d.shapes", "Box")
lazy_import("sage.plot.plot3d.texture", "Texture")

# ###################### predefined colors ##################

named_colors = {
    'red': (1, 0, 0),              # F face
    'green': (0, 1, 0),            # R face
    'blue': (0, 0, 1),             # D face
    'yellow': (1, 1, 0),           # L face
    'white': (1, 1, 1),            # none
    'orange': (1, 0.6, 0.3),       # B face
    'purple': (1, 0, 1),           # none
    'lpurple': (1, 0.63, 1),       # U face
    'lightblue': (0, 1, 1),        # none
    'lgrey': (0.75, 0.75, 0.75),   # sagemath.org color
}
globals().update(named_colors)

# ########################################################
# written by Tom Boothby, placed in the public domain


def xproj(x, y, z, r):
    r"""
    Return the `x`-projection of `(x,y,z)` rotated by `r`.

    EXAMPLES::

        sage: from sage.groups.perm_gps.cubegroup import rotation_list, xproj
        sage: rot = rotation_list(30, 45)                                               # needs sage.symbolic
        sage: xproj(1,2,3,rot)                                                          # needs sage.symbolic
        0.6123724356957945
    """
    return (y*r[1] - x*r[3])*r[2]


def yproj(x, y, z, r):
    r"""
    Return the `y`-projection of `(x,y,z)` rotated by `r`.

    EXAMPLES::

        sage: from sage.groups.perm_gps.cubegroup import rotation_list, yproj
        sage: rot = rotation_list(30, 45)                                               # needs sage.symbolic
        sage: yproj(1,2,3,rot)                                                          # needs sage.symbolic
        1.378497416975604
    """
    return z*r[2] - (x*r[1] + y*r[2])*r[0]


def rotation_list(tilt, turn):
    r"""
    Return a list `[\sin(\theta), \sin(\phi), \cos(\theta), \cos(\phi)]` of
    rotations where `\theta` is ``tilt`` and `\phi` is ``turn``.

    EXAMPLES::

        sage: from sage.groups.perm_gps.cubegroup import rotation_list
        sage: rotation_list(30, 45)                                                     # needs sage.symbolic
        [0.49999999999999994, 0.7071067811865475, 0.8660254037844387, 0.7071067811865476]
    """
    from sage.functions.trig import sin
    from sage.functions.trig import cos
    return [sin(tilt*pi/180.0), sin(turn*pi/180.0),
            cos(tilt*pi/180.0), cos(turn*pi/180.0)]


def polygon_plot3d(points, tilt=30, turn=30, **kwargs):
    r"""
    Plot a polygon viewed from an angle determined by ``tilt``, ``turn``, and
    vertices ``points``.

    .. WARNING::

       The ordering of the points is important to get "correct"
       and if you add several of these plots together, the one added first
       is also drawn first (ie, addition of Graphics objects is not
       commutative).

    The following example produced a green-colored square with vertices
    at the points indicated.

    EXAMPLES::

        sage: from sage.groups.perm_gps.cubegroup import polygon_plot3d,green
        sage: P = polygon_plot3d([[1,3,1],[2,3,1],[2,3,2],[1,3,2],[1,3,1]],             # needs sage.plot
        ....:                    rgbcolor=green)
    """
    rot = rotation_list(tilt, turn)
    points2 = [(xproj(x, y, z, rot), yproj(x, y, z, rot))
               for (x, y, z) in points]
    return polygon(points2, **kwargs)

# ##########################################################

# ############  lots of "internal" utility plot functions #########


def inv_list(lst):
    r"""
    Input a list of ints `1, \ldots, m` (in any order), outputs inverse
    perm.

    EXAMPLES::

        sage: from sage.groups.perm_gps.cubegroup import inv_list
        sage: L = [2,3,1]
        sage: inv_list(L)
        [3, 1, 2]
    """
    return [lst.index(i) + 1 for i in range(1, 1 + len(lst))]


face_polys = {
# ## bottom layer L, F, R, B
    'ldb': [[-3,0],[-2,0], [-2,1], [-3,1]],      # square labeled 14
    'ld': [[-2,0],[-1,0], [-1,1], [-2,1]],      # square labeled 15
    'lfd': [[-1,0],[0,0], [0,1], [-1,1]],      # square labeled 16
    'fdl': [[0,0],[1,0], [1,1], [0,1]],      # square labeled 22
    'fd': [[1,0],[2,0], [2,1], [1,1]],      # square labeled 23
    'frd': [[2,0],[3,0], [3,1], [2,1]],      # square labeled 24
    'rdf': [[3,0],[4,0], [4,1], [3,1]],      # square labeled 30
    'rd': [[4,0],[5,0], [5,1], [4,1]],      # square labeled 31
    'rbd': [[5,0],[6,0], [6,1], [5,1]],      # square labeled 32
    'bdr': [[6,0],[7,0], [7,1], [6,1]],      # square labeled 38
    'bd': [[7,0],[8,0], [8,1], [7,1]],      # square labeled 39
    'bld': [[8,0],[9,0], [9,1], [8,1]],      # square labeled 40
# ## middle layer L,F,R, B
    'lb': [[-3,1],[-2,1], [-2,2], [-3,2]],      # square labeled 12
    'l_center': [[-2,1],[-1,1], [-1,2], [-2,2]],        # center square
    'lf': [[-1,1],[0,1], [0,2], [-1,2]],      # square labeled 13
    'fl': [[0,1],[1,1], [1,2], [0,2]],      # square labeled 20
    'f_center': [[1,1],[2,1], [2,2], [1,2]],        # center square
    'fr': [[2,1],[3,1], [3,2], [2,2]],      # square labeled 21
    'rf': [[3,1],[4,1], [4,2], [3,2]],      # square labeled 28
    'r_center': [[4,1],[5,1], [5,2], [4,2]],        # center square
    'rb': [[5,1],[6,1], [6,2], [5,2]],      # square labeled 29
    'br': [[6,1],[7,1], [7,2], [6,2]],      # square labeled 36
    'b_center': [[7,1],[8,1], [8,2], [7,2]],        # center square
    'bl': [[8,1],[9,1], [9,2], [8,2]],      # square labeled 37
# ## top layer L, F, R, B
    'lbu': [[-3,2],[-2,2], [-2,3], [-3,3]],      # square labeled 9
    'lu': [[-2,2],[-1,2], [-1,3], [-2,3]],      # square labeled 10
    'luf': [[-1,2],[0,2], [0,3], [-1,3]],      # square labeled 11
    'flu': [[0,2],[1,2], [1,3], [0,3]],      # square labeled 17
    'fu': [[1,2],[2,2], [2,3], [1,3]],      # square labeled 18
    'fur': [[2,2],[3,2], [3,3], [2,3]],      # square labeled 19
    'ruf': [[3,2],[4,2], [4,3], [3,3]],      # square labeled 25
    'ru': [[4,2],[5,2], [5,3], [4,3]],      # square labeled 26
    'rub': [[5,2],[6,2], [6,3], [5,3]],      # square labeled 27
    'bur': [[6,2],[7,2], [7,3], [6,3]],      # square labeled 33
    'bu': [[7,2],[8,2], [8,3], [7,3]],      # square labeled 34
    'bul': [[8,2],[9,2], [9,3], [8,3]],      # square labeled 35
# down face
    'dlf': [[0,-1],[1,-1], [1,0], [0,0]],      # square labeled 41
    'df': [[1,-1],[2,-1], [2,0], [1,0]],      # square labeled 42
    'dfr': [[2,-1],[3,-1], [3,0], [2,0]],      # square labeled 43
    'dl': [[0,-2],[1,-2], [1,-1], [0,-1]],      # square labeled 44
    'd_center': [[1,-2],[2,-2], [2,-1], [1,-1]],        # center square
    'dr': [[2,-2],[3,-2], [3,-1], [2,-1]],      # square labeled 45
    'dlb': [[0,-3],[1,-3], [1,-2], [0,-2]],      # square labeled 46
    'db': [[1,-3],[2,-3], [2,-2], [1,-2]],      # square labeled 47
    'drb': [[2,-3],[3,-3], [3,-2], [2,-2]],      # square labeled 48
# up face
    'ufl': [[0,3],[1,3], [1,4], [0,4]],      # square labeled 6
    'uf': [[1,3],[2,3], [2,4], [1,4]],      # square labeled 7
    'urf': [[2,3],[3,3], [3,4], [2,4]],      # square labeled 8
    'ul': [[0,4],[1,4], [1,5], [0,5]],      # square labeled 4
    'u_center': [[1,4],[2,4], [2,5], [1,5]],        # center square
    'ur': [[2,4],[3,4], [3,5], [2,5]],      # square labeled 5
    'ulb': [[0,6],[1,6], [1,5], [0,5]],      # square labeled 1
    'ub': [[1,6],[2,6], [2,5], [1,5]],      # square labeled 2
    'ubr': [[2,6],[3,6], [3,5], [2,5]],      # square labeled 3
}


def create_poly(face, color):
    """
    Create the polygon given by ``face`` with color ``color``.

    EXAMPLES::

        sage: from sage.groups.perm_gps.cubegroup import create_poly, red
        sage: create_poly('ur', red)                                                    # needs sage.plot
        Graphics object consisting of 1 graphics primitive
    """
    return polygon(face_polys[face], rgbcolor=color)

####################################################


singmaster_indices = {
    1: "ulb",
    2: "ub",
    3: "ubr",
    4: "ul",
    5: "ur",
    6: "ufl",
    7: "uf",
    8: "urf",
    14: "ldb",
    15: "ld",
    16: "lfd",
    12: "lb",
    13: "lf",
    9: "lbu",
    10: "lu",
    11: "luf",
    17: "flu",
    18: "fu",
    19: "fur",
    20: "fl",
    21: "fr",
    22: "fdl",
    23: "fd",
    24: "frd",
    41: "dlf",
    42: "df",
    43: "dfr",
    44: "dl",
    45: "dr",
    46: "dlb",
    47: "db",
    48: "drb",
    33: "bur",
    34: "bu",
    35: "bul",
    36: "br",
    37: "bl",
    38: "bdr",
    39: "bd",
    40: "bld",
    25: "ruf",
    26: "ru",
    27: "rub",
    28: "rf",
    29: "rb",
    30: "rdf",
    31: "rd",
    32: "rbd",
}


def index2singmaster(facet):
    """
    Translate index used (eg, 43) to Singmaster facet notation (eg,
    fdr).

    EXAMPLES::

        sage: from sage.groups.perm_gps.cubegroup import index2singmaster
        sage: index2singmaster(41)
        'dlf'
    """
    return singmaster_indices[facet]


def color_of_square(facet, colors=['lpurple', 'yellow', 'red', 'green', 'orange', 'blue']):
    """
    Return the color the facet has in the solved state.

    EXAMPLES::

        sage: from sage.groups.perm_gps.cubegroup import color_of_square
        sage: color_of_square(41)
        'blue'
    """
    return colors[(facet - 1) // 8]


cubie_center_list = {
    #  centers of the cubies on the F,U, R faces
    1: [1//2, 1//2, 5//2],  # ulb
    2: [1//2, 3//2, 5//2],  #  ub
    3: [1//2, 5//2, 5//2],  #  ubr
    4: [3//2, 1//2, 5//2],  #  ul
    5: [3//2, 5//2, 5//2],  #  ur
    6: [5//2, 1//2, 5//2],  #  ufl
    7: [5//2, 3//2, 5//2],  #  uf
    8: [5//2, 5//2, 5//2],  #  urf
    17: [5//2, 1//2, 5//2],  #  flu
    18: [5//2, 3//2, 5//2],  #  fu
    19: [5//2, 5//2, 5//2],  #  fur
    20: [5//2, 1//2, 3//2],  #  fl
    21: [5//2, 5//2, 3//2],  #  fr
    22: [5//2, 1//2, 1//2],  #  fdl
    23: [5//2, 3//2, 1//2],  #  fd
    24: [5//2, 5//2, 1//2],  #  frd
    25: [5//2, 5//2, 5//2],  # rfu
    26: [3//2, 5//2, 5//2],  #  ru
    27: [1//2, 5//2, 5//2],  #  rub
    28: [5//2, 5//2, 3//2],  #  rf
    29: [1//2, 5//2, 3//2],  #  rb
    30: [5//2, 5//2, 1//2],  #  rdf
    31: [3//2, 5//2, 1//2],  #  rd
    32: [1//2, 5//2, 1//2],  # rbd
}


def cubie_centers(label):
    r"""
    Return the cubie center list element given by ``label``.

    EXAMPLES::

        sage: from sage.groups.perm_gps.cubegroup import cubie_centers
        sage: cubie_centers(3)
        [0, 2, 2]
    """
    return cubie_center_list[label]


def cubie_colors(label, state0):
    r"""
    Return the color of the cubie given by ``label`` at ``state0``.

    EXAMPLES::

        sage: from sage.groups.perm_gps.cubegroup import cubie_colors
        sage: G = CubeGroup()
        sage: g = G.parse("R*U")
        sage: cubie_colors(3, G.facets(g))
        [(1, 1, 1), (1, 0.63, 1), (1, 0.6, 0.3)]
    """
    #  colors of the cubies on the F,U, R faces
    clr_any = named_colors['white']
    state = inv_list(state0)
    if label == 1:
        return [clr_any, named_colors[color_of_square(state[1-1])], clr_any] #ulb,
    if label == 2:
        return [clr_any,named_colors[color_of_square(state[2-1])],clr_any] # ub,
    if label == 3:
        return [clr_any, named_colors[color_of_square(state[3-1])], named_colors[color_of_square(state[27-1])]] # ubr,
    if label == 4:
        return [clr_any, named_colors[color_of_square(state[4-1])], clr_any] # ul,
    if label == 5:
        return [clr_any, named_colors[color_of_square(state[5-1])], named_colors[color_of_square(state[26-1])]] # ur,
    if label == 6:
        return [named_colors[color_of_square(state[17-1])], named_colors[color_of_square(state[6-1])], clr_any] # ufl,
    if label == 7:
        return [named_colors[color_of_square(state[18-1])], named_colors[color_of_square(state[7-1])], clr_any] # uf,
    if label == 8:
        return [named_colors[color_of_square(state[19-1])], named_colors[color_of_square(state[8-1])], named_colors[color_of_square(state[25-1])]] # urf,
    if label == 17:
        return [named_colors[color_of_square(state[17-1])], named_colors[color_of_square(state[6-1])], clr_any] # flu
    if label == 18:
        return [named_colors[color_of_square(state[18-1])], named_colors[color_of_square(state[7-1])], clr_any] # fu
    if label == 19:
        return [named_colors[color_of_square(state[19-1])], named_colors[color_of_square(state[8-1])], named_colors[color_of_square(state[25-1])]] # fur
    if label == 20:
        return [named_colors[color_of_square(state[20-1])], clr_any, clr_any] # fl
    if label == 21:
        return [named_colors[color_of_square(state[21-1])], clr_any, named_colors[color_of_square(state[28-1])]] # fr
    if label == 22:
        return [named_colors[color_of_square(state[22-1])], clr_any, clr_any] # fdl
    if label == 23:
        return [named_colors[color_of_square(state[23-1])], clr_any, clr_any] # fd
    if label == 24:
        return [named_colors[color_of_square(state[24-1])], clr_any, named_colors[color_of_square(state[30-1])]] # frd
    if label == 25:
        return [named_colors[color_of_square(state[19-1])],named_colors[color_of_square(state[8-1])],named_colors[color_of_square(state[25-1])]]  #rfu,
    if label == 26:
        return [clr_any,named_colors[color_of_square(state[5-1])],named_colors[color_of_square(state[26-1])]] # ru,
    if label == 27:
        return [clr_any,named_colors[color_of_square(state[3-1])],named_colors[color_of_square(state[27-1])]] # rub,
    if label == 28:
        return [named_colors[color_of_square(state[21-1])],clr_any,named_colors[color_of_square(state[28-1])]] # rf,
    if label == 29:
        return [clr_any,clr_any,named_colors[color_of_square(state[29-1])]] # rb,
    if label == 30:
        return [named_colors[color_of_square(state[24-1])],clr_any,named_colors[color_of_square(state[30-1])]] # rdf,
    if label == 31:
        return [clr_any,clr_any,named_colors[color_of_square(state[31-1])]] # rd,
    if label == 32:
        return [clr_any,clr_any,named_colors[color_of_square(state[32-1])]] #rbd,


def plot3d_cubie(cnt, clrs):
    r"""
    Plot the front, up and right face of a cubie centered at cnt and
    rgbcolors given by clrs (in the order FUR).

    Type ``P.show()`` to view.

    EXAMPLES::

        sage: from sage.groups.perm_gps.cubegroup import plot3d_cubie, blue, red, green
        sage: clrF = blue; clrU = red; clrR = green
        sage: P = plot3d_cubie([1/2,1/2,1/2],[clrF,clrU,clrR])                          # needs sage.plot
    """
    half = QQ((1, 2))
    x = cnt[0] - half
    y = cnt[1] - half
    z = cnt[2] - half
    #ptsD = [[x+0,y+0,0+z],[x+1,y+0,0+z],[x+1,y+1,0+z],[x+0,y+1,0+z],[x+0,y+0,0+z]]
    ptsF = [[x+1,y+0,0+z],[x+1,y+1,0+z],[x+1,y+1,1+z],[x+1,y+0,1+z],[x+1,y+0,0+z]]
    #ptsB = [[x+0,y+0,0+z],[x+0,y+1,0+z],[x+0,y+1,1+z],[x+0,y+0,1+z],[x+0,y+0,0+z]]
    ptsU = [[x+0,y+0,1+z],[x+1,y+0,1+z],[x+1,y+1,1+z],[x+0,y+1,1+z],[x+0,y+0,1+z]]
    #ptsL = [[x+0,y+0,0+z],[x+1,y+0,0+z],[x+1,y+0,1+z],[x+0,y+0,1+z],[x+0,y+0,0+z]]
    ptsR = [[x+0,y+1,0+z],[x+1,y+1,0+z],[x+1,y+1,1+z],[x+0,y+1,1+z],[x+0,y+1,0+z]]
    P = polygon_plot3d(ptsR, rgbcolor=clrs[2])
    P += polygon_plot3d(ptsU, rgbcolor=clrs[1])
    P += polygon_plot3d(ptsF, rgbcolor=clrs[0])
    P.axes(show=False)
    return P


# ################### end of "internal" utility plot functions  ########


class CubeGroup(PermutationGroup_generic):
    r"""
    A python class to help compute Rubik's cube group actions.

    .. NOTE::

        This group is also available via ``groups.permutation.RubiksCube()``.

    EXAMPLES:

    If G denotes the cube group then it may be regarded as a
    subgroup of ``SymmetricGroup(48)``, where the 48 facets are labeled as
    follows.

    ::

        sage: rubik = CubeGroup()
        sage: rubik.display2d("")
                     ┌──────────────┐
                     │  1    2    3 │
                     │  4   top   5 │
                     │  6    7    8 │
        ┌────────────┼──────────────┼─────────────┬────────────┐
        │  9  10  11 │ 17   18   19 │ 25   26  27 │ 33  34  35 │
        │ 12 left 13 │ 20  front 21 │ 28 right 29 │ 36 rear 37 │
        │ 14  15  16 │ 22   23   24 │ 30   31  32 │ 38  39  40 │
        └────────────┼──────────────┼─────────────┴────────────┘
                     │ 41   42   43 │
                     │ 44 bottom 45 │
                     │ 46   47   48 │
                     └──────────────┘

    ::

        sage: rubik
        The Rubik's cube group with generators R,L,F,B,U,D in SymmetricGroup(48).

        TESTS::

            sage: groups.permutation.RubiksCube()
            The Rubik's cube group with generators R,L,F,B,U,D in SymmetricGroup(48).
    """
    def __init__(self):
        """
        Initialize ``self``.

        EXAMPLES::

            sage: rubik = CubeGroup()
            sage: TestSuite(rubik).run(skip='_test_enumerated_set_contains') # because the group is very large

        TESTS:

        Check that :issue:`11360` is fixed::

            sage: rubik = CubeGroup()
            sage: rubik.order()
            43252003274489856000
        """
        U = "( 1, 3, 8, 6)( 2, 5, 7, 4)( 9,33,25,17)(10,34,26,18)(11,35,27,19)"  # U = top
        L = "( 9,11,16,14)(10,13,15,12)( 1,17,41,40)( 4,20,44,37)( 6,22,46,35)"  # L = left
        F = "(17,19,24,22)(18,21,23,20)( 6,25,43,16)( 7,28,42,13)( 8,30,41,11)"  # F = front
        R = "(25,27,32,30)(26,29,31,28)( 3,38,43,19)( 5,36,45,21)( 8,33,48,24)"  # R = right
        B = "(33,35,40,38)(34,37,39,36)( 3, 9,46,32)( 2,12,47,29)( 1,14,48,27)"  # B = back or rear
        D = "(41,43,48,46)(42,45,47,44)(14,22,30,38)(15,23,31,39)(16,24,32,40)"  # D = down or bottom
        PermutationGroup_generic.__init__(self, gens=[B, D, F, L, R, U], canonicalize=False)

    def gen_names(self):
        """
        Return the names of the generators.

        EXAMPLES::

            sage: rubik = CubeGroup()
            sage: rubik.gen_names()
            ['B', 'D', 'F', 'L', 'R', 'U']
        """
        return ['B','D','F','L','R','U']

    def __repr__(self):
        """
        Return a string representation of ``self``.

        EXAMPLES::

            sage: rubik = CubeGroup()
            sage: rubik
            The Rubik's cube group with generators R,L,F,B,U,D in SymmetricGroup(48).
        """
        return "The Rubik's cube group with generators R,L,F,B,U,D in SymmetricGroup(48)."

    def B(self):
        """
        Return the generator `B` in Singmaster notation.

        EXAMPLES::

            sage: rubik = CubeGroup()
            sage: rubik.B()
            (1,14,48,27)(2,12,47,29)(3,9,46,32)(33,35,40,38)(34,37,39,36)
        """
        return self.gens()[0]

    def D(self):
        """
        Return the generator `D` in Singmaster notation.

        EXAMPLES::

            sage: rubik = CubeGroup()
            sage: rubik.D()
            (14,22,30,38)(15,23,31,39)(16,24,32,40)(41,43,48,46)(42,45,47,44)
        """
        return self.gens()[1]

    def F(self):
        """
        Return the generator `F` in Singmaster notation.

        EXAMPLES::

            sage: rubik = CubeGroup()
            sage: rubik.F()
            (6,25,43,16)(7,28,42,13)(8,30,41,11)(17,19,24,22)(18,21,23,20)
        """
        return self.gens()[2]

    def L(self):
        """
        Return the generator `L` in Singmaster notation.

        EXAMPLES::

            sage: rubik = CubeGroup()
            sage: rubik.L()
            (1,17,41,40)(4,20,44,37)(6,22,46,35)(9,11,16,14)(10,13,15,12)
        """
        return self.gens()[3]

    def R(self):
        """
        Return the generator `R` in Singmaster notation.

        EXAMPLES::

            sage: rubik = CubeGroup()
            sage: rubik.R()
            (3,38,43,19)(5,36,45,21)(8,33,48,24)(25,27,32,30)(26,29,31,28)
        """
        return self.gens()[4]

    def U(self):
        """
        Return the generator `U` in Singmaster notation.

        EXAMPLES::

            sage: rubik = CubeGroup()
            sage: rubik.U()
            (1,3,8,6)(2,5,7,4)(9,33,25,17)(10,34,26,18)(11,35,27,19)
        """
        return self.gens()[5]

    def parse(self, mv, check=True):
        r"""
        This function allows one to create the permutation group element
        from a variety of formats.

        INPUT:

        - ``mv`` -- can one of the following:

          - ``list`` -- list of facets (as returned by
            self.facets())

          - ``dict`` -- list of faces (as returned by
            ``self.faces()``)

          - ``str`` -- either cycle notation (passed to GAP) or
            a product of generators or Singmaster notation

          - ``perm_group element`` -- returned as an element of ``self``

        - ``check`` -- check if the input is valid

        EXAMPLES::

            sage: C = CubeGroup()
            sage: C.parse(list(range(1,49)))
            ()
            sage: g = C.parse("L"); g
            (1,17,41,40)(4,20,44,37)(6,22,46,35)(9,11,16,14)(10,13,15,12)
            sage: C.parse(str(g)) == g
            True
            sage: facets = C.facets(g); facets
            [17, 2, 3, 20, 5, 22, 7, 8, 11, 13, 16, 10, 15, 9, 12, 14, 41, 18, 19, 44, 21, 46, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 6, 36, 4, 38, 39, 1, 40, 42, 43, 37, 45, 35, 47, 48]
            sage: C.parse(facets)
            (1,17,41,40)(4,20,44,37)(6,22,46,35)(9,11,16,14)(10,13,15,12)
            sage: C.parse(facets) == g
            True
            sage: faces = C.faces("L"); faces
            {'back': [[33, 34, 6], [36, 0, 4], [38, 39, 1]],
             'down': [[40, 42, 43], [37, 0, 45], [35, 47, 48]],
             'front': [[41, 18, 19], [44, 0, 21], [46, 23, 24]],
             'left': [[11, 13, 16], [10, 0, 15], [9, 12, 14]],
             'right': [[25, 26, 27], [28, 0, 29], [30, 31, 32]],
             'up': [[17, 2, 3], [20, 0, 5], [22, 7, 8]]}
            sage: C.parse(faces) == C.parse("L")
            True
            sage: C.parse("L' R2") == C.parse("L^(-1)*R^2")
            True
            sage: C.parse("L' R2")
            (1,40,41,17)(3,43)(4,37,44,20)(5,45)(6,35,46,22)(8,48)(9,14,16,11)(10,12,15,13)(19,38)(21,36)(24,33)(25,32)(26,31)(27,30)(28,29)
            sage: C.parse("L^4")
            ()
            sage: C.parse("L^(-1)*R")
            (1,40,41,17)(3,38,43,19)(4,37,44,20)(5,36,45,21)(6,35,46,22)(8,33,48,24)(9,14,16,11)(10,12,15,13)(25,27,32,30)(26,29,31,28)
        """
        if isinstance(mv, PermutationGroupElement):
            # mv is a perm_group element, return mv
            return mv if mv.parent() is self else PermutationGroup_generic.__call__(self, mv, check)
        elif isinstance(mv, str):
            # It is a string: may be in cycle notation or Rubik's notation
            if '(' in mv and '^' not in mv:
                return PermutationGroup_generic.__call__(self, mv, check)
            else:
                gens = self.gens()
                names = self.gen_names()
                map = {}
                for i in range(6):
                    map[names[i]] = gens[i]
                g = self.identity()
                mv = mv.strip().replace(" ","*").replace("**", "*").replace("'", "-1").replace('^','').replace('(','').replace(')','')
                M = mv.split("*")
                for m in M:
                    if not m:
                        pass
                    elif len(m) == 1:
                        g *= map[m[0]]
                    else:
                        g *= map[m[0]]**int(m[1:])
                return g
        elif isinstance(mv, dict):
            state = mv
            state_facets = []
            keyss = sorted(state.keys())
            for k in keyss:
                r = state[k][0]+state[k][1]+state[k][2]
                r.remove(0)
                state_facets = state_facets + r
            state0 = self.faces("")
            state0_facets = []
            keyss = sorted(state0.keys())
            for k in keyss:
                r = state0[k][0]+state0[k][1]+state0[k][2]
                r.remove(0)
                state0_facets = state0_facets + r
            p1 = [state0_facets.index(x) for x in range(1,49)]
            p2 = [state_facets[j] for j in p1]
            return PermutationGroup_generic.__call__(self, p2, check)
        else:
            return PermutationGroup_generic.__call__(self, mv, check)

    __call__ = parse

    def facets(self, g=None):
        r"""
        Return the set of facets on which the group acts. This function is
        a "constant".

        EXAMPLES::

            sage: rubik = CubeGroup()
            sage: rubik.facets() == list(range(1,49))
            True
        """
        fcts = range(1, 49)
        if g is not None:
            return [g(i) for i in fcts]
        else:
            return list(fcts)

    def faces(self, mv):
        r"""
        Return the dictionary of faces created by the effect of the move
        mv, which is a string of the form `X^a*Y^b*...`, where
        `X, Y, \ldots` are in `\{R,L,F,B,U,D\}` and
        `a, b, \ldots` are integers. We call this ordering of the faces
        the "BDFLRU, L2R, T2B ordering".

        EXAMPLES::

            sage: rubik = CubeGroup()

        Here is the dictionary of the solved state::

            sage: sorted(rubik.faces("").items())
            [('back', [[33, 34, 35], [36, 0, 37], [38, 39, 40]]),
             ('down', [[41, 42, 43], [44, 0, 45], [46, 47, 48]]),
             ('front', [[17, 18, 19], [20, 0, 21], [22, 23, 24]]),
             ('left', [[9, 10, 11], [12, 0, 13], [14, 15, 16]]),
             ('right', [[25, 26, 27], [28, 0, 29], [30, 31, 32]]),
             ('up', [[1, 2, 3], [4, 0, 5], [6, 7, 8]])]

        Now the dictionary of the state obtained after making the move `R`
        followed by `L`::

            sage: sorted(rubik.faces("R*U").items())
            [('back', [[48, 26, 27], [45, 0, 37], [43, 39, 40]]),
             ('down', [[41, 42, 11], [44, 0, 21], [46, 47, 24]]),
             ('front', [[9, 10, 8], [20, 0, 7], [22, 23, 6]]),
             ('left', [[33, 34, 35], [12, 0, 13], [14, 15, 16]]),
             ('right', [[19, 29, 32], [18, 0, 31], [17, 28, 30]]),
             ('up', [[3, 5, 38], [2, 0, 36], [1, 4, 25]])]
        """
        fcts = self.facets(self.parse(mv))
        faceR = [[fcts[24],fcts[25],fcts[26]],[fcts[27],0,fcts[28]],[fcts[29],fcts[30],fcts[31]]]
        faceL = [[fcts[8],fcts[9],fcts[10]],[fcts[11],0,fcts[12]],[fcts[13],fcts[14],fcts[15]]]
        faceU = [[fcts[0],fcts[1],fcts[2]],[fcts[3],0,fcts[4]],[fcts[5],fcts[6],fcts[7]]]
        faceD = [[fcts[40],fcts[41],fcts[42]],[fcts[43],0,fcts[44]],[fcts[45],fcts[46],fcts[47]]]
        faceF = [[fcts[16],fcts[17],fcts[18]],[fcts[19],0,fcts[20]],[fcts[21],fcts[22],fcts[23]]]
        faceB = [[fcts[32],fcts[33],fcts[34]],[fcts[35],0,fcts[36]],[fcts[37],fcts[38],fcts[39]]]
        return {'right': faceR, 'left': faceL, 'up': faceU, 'down': faceD, 'front': faceF, 'back': faceB}

    def move(self, mv):
        r"""
        Return the group element and the reordered list of facets, as
        moved by the list ``mv`` (read left-to-right)

        INPUT:

        - ``mv`` -- string of the form ``Xa*Yb*...``,
          where ``X``, ``Y``, ... are in ``R``, ``L``, ``F``, ``B``, ``U``,
          ``D`` and ``a``, ``b``, ... are integers

        EXAMPLES::

            sage: rubik = CubeGroup()
            sage: rubik.move("")[0]
            ()
            sage: rubik.move("R")[0]
            (3,38,43,19)(5,36,45,21)(8,33,48,24)(25,27,32,30)(26,29,31,28)
            sage: rubik.R()
            (3,38,43,19)(5,36,45,21)(8,33,48,24)(25,27,32,30)(26,29,31,28)
        """
        g = self.parse(mv)
        return g, self.facets(g)

    def display2d(self, mv):
        r"""
        Print the 2d representation of ``self``.

        EXAMPLES::

            sage: rubik = CubeGroup()
            sage: rubik.display2d("R")
                         ┌──────────────┐
                         │  1    2   38 │
                         │  4   top  36 │
                         │  6    7   33 │
            ┌────────────┼──────────────┼─────────────┬────────────┐
            │  9  10  11 │ 17   18    3 │ 27   29  32 │ 48  34  35 │
            │ 12 left 13 │ 20  front  5 │ 26 right 31 │ 45 rear 37 │
            │ 14  15  16 │ 22   23    8 │ 25   28  30 │ 43  39  40 │
            └────────────┼──────────────┼─────────────┴────────────┘
                         │ 41   42   19 │
                         │ 44 bottom 21 │
                         │ 46   47   24 │
                         └──────────────┘
        """
        print(self.repr2d(mv))

    def repr2d(self, mv):
        r"""
        Displays a 2D map of the Rubik's cube after the move mv has been
        made. Nothing is returned.

        EXAMPLES::

            sage: rubik = CubeGroup()
            sage: print(rubik.repr2d(""))
                         ┌──────────────┐
                         │  1    2    3 │
                         │  4   top   5 │
                         │  6    7    8 │
            ┌────────────┼──────────────┼─────────────┬────────────┐
            │  9  10  11 │ 17   18   19 │ 25   26  27 │ 33  34  35 │
            │ 12 left 13 │ 20  front 21 │ 28 right 29 │ 36 rear 37 │
            │ 14  15  16 │ 22   23   24 │ 30   31  32 │ 38  39  40 │
            └────────────┼──────────────┼─────────────┴────────────┘
                         │ 41   42   43 │
                         │ 44 bottom 45 │
                         │ 46   47   48 │
                         └──────────────┘

        ::

            sage: print(rubik.repr2d("R"))
                         ┌──────────────┐
                         │  1    2   38 │
                         │  4   top  36 │
                         │  6    7   33 │
            ┌────────────┼──────────────┼─────────────┬────────────┐
            │  9  10  11 │ 17   18    3 │ 27   29  32 │ 48  34  35 │
            │ 12 left 13 │ 20  front  5 │ 26 right 31 │ 45 rear 37 │
            │ 14  15  16 │ 22   23    8 │ 25   28  30 │ 43  39  40 │
            └────────────┼──────────────┼─────────────┴────────────┘
                         │ 41   42   19 │
                         │ 44 bottom 21 │
                         │ 46   47   24 │
                         └──────────────┘

        You can see the right face has been rotated but not the left face.
        """
        g = self.parse(mv)
        lst = self.facets(g)
        line1 = "             ┌──────────────┐\n"
        line2 = "             │%3d  %3d  %3d │\n" % (lst[0], lst[1], lst[2])
        line3 = "             │%3d   top %3d │\n" % (lst[3], lst[4])
        line4 = "             │%3d  %3d  %3d │\n" % (lst[5], lst[6], lst[7])
        line5 = "┌────────────┼──────────────┼─────────────┬────────────┐\n"
        line6 = "│%3d %3d %3d │%3d  %3d  %3d │%3d  %3d %3d │%3d %3d %3d │\n" % (lst[8], lst[9], lst[10], lst[16], lst[17], lst[18], lst[24], lst[25], lst[26], lst[32], lst[33], lst[34])
        line7 = "│%3d left%3d │%3d  front%3d │%3d right%3d │%3d rear%3d │\n" % (lst[11], lst[12], lst[19], lst[20], lst[27], lst[28], lst[35], lst[36])
        line8 = "│%3d %3d %3d │%3d  %3d  %3d │%3d  %3d %3d │%3d %3d %3d │\n" % (lst[13], lst[14], lst[15], lst[21], lst[22], lst[23], lst[29], lst[30], lst[31], lst[37], lst[38], lst[39])
        line9 = "└────────────┼──────────────┼─────────────┴────────────┘\n"
        line10 = "             │%3d  %3d  %3d │\n" % (lst[40], lst[41], lst[42])
        line11 = "             │%3d bottom%3d │\n" % (lst[43], lst[44])
        line12 = "             │%3d  %3d  %3d │\n" % (lst[45], lst[46], lst[47])
        line13 = "             └──────────────┘\n"
        return line1+line2+line3+line4+line5+line6+line7+line8+line9+line10+line11+line12+line13

    def plot_cube(self, mv, title=True, colors=[lpurple, yellow, red, green, orange, blue]):
        r"""
        Input the move mv, as a string in the Singmaster notation, and
        output the 2D plot of the cube in that state.

        Type ``P.show()`` to display any of the plots below.

        EXAMPLES::

            sage: rubik = CubeGroup()
            sage: P = rubik.plot_cube("R^2*U^2*R^2*U^2*R^2*U^2", title=False)           # needs sage.plot
            sage: # (R^2U^2)^3  permutes 2 pairs of edges (uf,ub)(fr,br)
            sage: P = rubik.plot_cube("R*L*D^2*B^3*L^2*F^2*R^2*U^3*D*R^3*D^2*F^3*B^3*D^3*F^2*D^3*R^2*U^3*F^2*D^3")      # needs sage.plot
            sage: # the superflip (in 20f* moves)
            sage: P = rubik.plot_cube("U^2*F*U^2*L*R^(-1)*F^2*U*F^3*B^3*R*L*U^2*R*D^3*U*L^3*R*D*R^3*L^3*D^2")           # needs sage.plot
            sage: # "superflip+4 spot" (in 26q* moves)
        """
        g = self.parse(mv)
        state = self.facets(g)
        cubies = [create_poly(index2singmaster(state[x]), color_of_square(x+1, colors)) for x in range(48)]
        centers = [create_poly('%s_center' % "ulfrbd"[i], colors[i]) for i in range(6)]
        clrs = sum(cubies) + sum(centers)
        clrs.axes(show=False)
        if title:
            t = text('sagemath.org', (7.8, -3.5),rgbcolor=lgrey)
            P = clrs + t
            P.axes(show=False)
            return P
        return clrs

    def plot3d_cube(self, mv, title=True):
        r"""
        Displays `F,U,R` faces of the cube after the given move ``mv``. Mostly
        included for the purpose of drawing pictures and checking moves.

        INPUT:

        - ``mv`` -- string in the Singmaster notation
        - ``title`` -- boolean (default: ``True``); display the title information

        The first one below is "superflip+4 spot" (in 26q\* moves) and the
        second one is the superflip (in 20f\* moves). Type show(P) to view
        them.

        EXAMPLES::

            sage: rubik = CubeGroup()
            sage: P = rubik.plot3d_cube("U^2*F*U^2*L*R^(-1)*F^2*U*F^3*B^3*R*L*U^2*R*D^3*U*L^3*R*D*R^3*L^3*D^2")         # needs sage.plot
            sage: P = rubik.plot3d_cube("R*L*D^2*B^3*L^2*F^2*R^2*U^3*D*R^3*D^2*F^3*B^3*D^3*F^2*D^3*R^2*U^3*F^2*D^3")    # needs sage.plot
        """
        g = self.parse(mv)
        state = self.facets(g)
        cubiesR = [plot3d_cubie(cubie_centers(c),cubie_colors(c,state)) for c in [32,31,30,29,28,27,26,25]]
        cubeR = sum(cubiesR)
        cubiesU = [plot3d_cubie(cubie_centers(c),cubie_colors(c,state)) for c in range(1,9)]
        cubeU = sum(cubiesU)
        cubiesF = [plot3d_cubie(cubie_centers(c),cubie_colors(c,state)) for c in [22,23,24,20,21]]
        cubeF = sum(cubiesF)
        centerR = polygon_plot3d([[1,3,1],[2,3,1],[2,3,2],[1,3,2],[1,3,1]],rgbcolor=green)
        centerF = polygon_plot3d([[3,1,1],[3,2,1],[3,2,2],[3,1,2],[3,1,1]],rgbcolor=red)
        centerU = polygon_plot3d([[1,1,3],[1,2,3],[2,2,3],[2,1,3],[1,1,3]],rgbcolor=lpurple)
        centers = centerF+centerR+centerU
        P = cubeR+cubeF+cubeU+centers
        P.axes(show=False)
        if title:
            t1 = text('Up, Front, and Right faces. ', (-0.2, -2.5))
            t2 = text('      sagemath.org', (0.8, -3.1), rgbcolor=lgrey)
            t3 = text("     ", (3.5, 0), rgbcolor=white)
            P = P + t1 + t2 + t3
            P.axes(show=False)
            return P
        return P

    def legal(self, state, mode='quiet'):
        r"""
        Return 1 (true) if the dictionary ``state`` (in the
        same format as returned by the faces method) represents a legal
        position (or state) of the Rubik's cube or 0 (false)
        otherwise.

        EXAMPLES::

            sage: rubik = CubeGroup()
            sage: r0 = rubik.faces("")
            sage: r1 = {'back': [[33, 34, 35], [36, 0, 37], [38, 39, 40]], 'down': [[41, 42, 43], [44, 0, 45], [46, 47, 48]],'front': [[17, 18, 19], [20, 0, 21], [22, 23, 24]],'left': [[9, 10, 11], [12, 0, 13], [14, 15, 16]],'right': [[25, 26, 27], [28, 0, 29], [30, 31, 32]],'up': [[1, 2, 3], [4, 0, 5], [6, 8, 7]]}
            sage: rubik.legal(r0)
            1
            sage: rubik.legal(r0,"verbose")
            (1, ())
            sage: rubik.legal(r1)
            0
        """
        try:
            g = self.parse(state)
            res = 1
        except ValueError:
            res = 0
            g = self.identity()
        if mode != 'quiet':
            return res, g
        else:
            return res

    def solve(self, state, algorithm='default'):
        r"""
        Solve the cube in the ``state``, given as a dictionary
        as in ``legal``. See the ``solve`` method
        of the RubiksCube class for more details.

        This may use GAP's ``EpimorphismFromFreeGroup`` and
        ``PreImagesRepresentative`` as explained below, if
        'gap' is passed in as the algorithm.

        This algorithm


        #. constructs the free group on 6 generators then computes a
           reasonable set of relations which they satisfy

        #. computes a homomorphism from the cube group to this free group
           quotient

        #. takes the cube position, regarded as a group element, and maps
           it over to the free group quotient

        #. using those relations and tricks from combinatorial group theory
           (stabilizer chains), solves the "word problem" for that element.

        #. uses python string parsing to rewrite that in cube notation.


        The Rubik's cube group has about `4.3 \times 10^{19}`
        elements, so this process is time-consuming. See
        https://www.gap-system.org/Doc/Examples/rubik.html for an
        interesting discussion of some GAP code analyzing the Rubik's
        cube.

        EXAMPLES::

            sage: rubik = CubeGroup()
            sage: state = rubik.faces("R")
            sage: rubik.solve(state)                                                    # optional - rubiks
            'R'
            sage: state = rubik.faces("R*U")
            sage: rubik.solve(state, algorithm='gap')       # long time
            'R*U'

        You can also check this another (but similar) way using the
        ``word_problem`` method (eg, G = rubik.group(); g =
        G("(3,38,43,19)(5,36,45,21)(8,33,48,24)(25,27,32,30)(26,29,31,28)");
        g.word_problem([b,d,f,l,r,u]), though the output will be less
        intuitive).
        """
        try:
            g = self.parse(state)
        except TypeError:
            return "Illegal or syntactically incorrect state. No solution."
        if algorithm != 'gap':
            C = RubiksCube(g)
            return C.solve(algorithm)

        group = libgap(self)
        hom = group.EpimorphismFromFreeGroup()
        sol = str(hom.PreImagesRepresentative(libgap(g)))
        for i, name in enumerate(self.gen_names()):
            sol = sol.replace("x%s" % (i + 1), name)
        return sol


##########################################################
#              3d object generation
##########################################################

def cubie_faces():
    """
    This provides a map from the 6 faces of the 27 cubies to the 48
    facets of the larger cube.

    -1,-1,-1 is left, top, front

    EXAMPLES::

        sage: from sage.groups.perm_gps.cubegroup import cubie_faces
        sage: sorted(cubie_faces().items())
        [((-1, -1, -1), [6, 17, 11, 0, 0, 0]),
         ((-1, -1, 0), [4, 0, 10, 0, 0, 0]),
         ((-1, -1, 1), [1, 0, 9, 0, 35, 0]),
         ((-1, 0, -1), [0, 20, 13, 0, 0, 0]),
         ((-1, 0, 0), [0, 0, -5, 0, 0, 0]),
         ((-1, 0, 1), [0, 0, 12, 0, 37, 0]),
         ((-1, 1, -1), [0, 22, 16, 41, 0, 0]),
         ((-1, 1, 0), [0, 0, 15, 44, 0, 0]),
         ((-1, 1, 1), [0, 0, 14, 46, 40, 0]),
         ((0, -1, -1), [7, 18, 0, 0, 0, 0]),
         ((0, -1, 0), [-6, 0, 0, 0, 0, 0]),
         ((0, -1, 1), [2, 0, 0, 0, 34, 0]),
         ((0, 0, -1), [0, -4, 0, 0, 0, 0]),
         ((0, 0, 0), [0, 0, 0, 0, 0, 0]),
         ((0, 0, 1), [0, 0, 0, 0, -2, 0]),
         ((0, 1, -1), [0, 23, 0, 42, 0, 0]),
         ((0, 1, 0), [0, 0, 0, -1, 0, 0]),
         ((0, 1, 1), [0, 0, 0, 47, 39, 0]),
         ((1, -1, -1), [8, 19, 0, 0, 0, 25]),
         ((1, -1, 0), [5, 0, 0, 0, 0, 26]),
         ((1, -1, 1), [3, 0, 0, 0, 33, 27]),
         ((1, 0, -1), [0, 21, 0, 0, 0, 28]),
         ((1, 0, 0), [0, 0, 0, 0, 0, -3]),
         ((1, 0, 1), [0, 0, 0, 0, 36, 29]),
         ((1, 1, -1), [0, 24, 0, 43, 0, 30]),
         ((1, 1, 0), [0, 0, 0, 45, 0, 31]),
         ((1, 1, 1), [0, 0, 0, 48, 38, 32])]
    """
    faceR = [[25, 26, 27], [28, -3, 29], [30, 31, 32]]  # green
    faceL = [[9, 10, 11], [12, -5, 13], [14, 15, 16]]   # orange
    faceU = [[1, 2, 3], [4, -6, 5], [6, 7, 8]]          # red
    faceD = [[41, 42, 43], [44, -1, 45], [46, 47, 48]]  # purple
    faceF = [[17, 18, 19], [20, -4, 21], [22, 23, 24]]  # yellow
    faceB = [[33, 34, 35], [36, -2, 37], [38, 39, 40]]  # blue
    cubies = {}
    for x in [-1, 0, 1]:
        for y in [-1, 0, 1]:
            for z in [-1, 0, 1]:
                cubies[x, y, z] = [0, 0, 0, 0, 0, 0]

    for i in [-1, 0, 1]:
        for j in [-1, 0, 1]:
            cubies[i, j, -1][1] = faceF[1 + j][1 + i]
            cubies[i, j, 1][4] = faceB[1 + j][1 - i]
            cubies[i, -1, j][0] = faceU[1 - j][1 + i]
            cubies[i, 1, j][3] = faceD[1 + j][1 + i]
            cubies[-1, i, j][2] = faceL[1 + i][1 - j]
            cubies[1, i, j][5] = faceR[1 + i][1 + j]

    return cubies


cubie_face_list = cubie_faces()


rand_colors = [(RDF.random_element(), RDF.random_element(), RDF.random_element()) for _ in range(56)]


@richcmp_method
class RubiksCube(SageObject):
    r"""
    The Rubik's cube (in a given state).

    EXAMPLES::

        sage: C = RubiksCube().move("R U R'")
        sage: C.show3d()                                                                # needs sage.plot

    ::

        sage: C = RubiksCube("R*L"); C
                     ┌──────────────┐
                     │ 17    2   38 │
                     │ 20   top  36 │
                     │ 22    7   33 │
        ┌────────────┼──────────────┼─────────────┬────────────┐
        │ 11  13  16 │ 41   18    3 │ 27   29  32 │ 48  34   6 │
        │ 10 left 15 │ 44  front  5 │ 26 right 31 │ 45 rear  4 │
        │  9  12  14 │ 46   23    8 │ 25   28  30 │ 43  39   1 │
        └────────────┼──────────────┼─────────────┴────────────┘
                     │ 40   42   19 │
                     │ 37 bottom 21 │
                     │ 35   47   24 │
                     └──────────────┘
        sage: C.show()                                                                  # needs sage.plot
        sage: C.solve(algorithm='gap')  # long time
        'L*R'
        sage: C == RubiksCube("L*R")
        True
    """
    def __init__(self, state=None, history=[], colors=[lpurple,yellow,red,green,orange,blue]):
        """
        Initialize ``self``.

        EXAMPLES::

            sage: C = RubiksCube().move("R*U")
            sage: TestSuite(C).run()
        """
        self.colors = colors
        self._history = history
        self._group = CubeGroup()
        if state is None:
            self._state = self._group.identity()
        else:
            if isinstance(state, str):
                state = self._group.faces(state)
            if not isinstance(state, PermutationGroupElement):
                legal, state = self._group.legal(state, mode='gimme_group_element')
                if not legal:
                    raise ValueError("not a legal cube")
            self._state = state

    def move(self, g):
        r"""
        Move the Rubik's cube by ``g``.

        EXAMPLES::

            sage: RubiksCube().move("R*U") == RubiksCube("R*U")
            True
        """
        if not isinstance(g, self._group.element_class):
            g = self._group.move(g)[0]
        return RubiksCube(self._state * g, self._history + [g], self.colors)

    def undo(self):
        r"""
        Undo the last move of the Rubik's cube.

        EXAMPLES::

            sage: C = RubiksCube()
            sage: D = C.move("R*U")
            sage: D.undo() == C
            True
        """
        if not self._history:
            raise ValueError("no moves to undo")
        g = self._history[-1]
        return RubiksCube(self._state * ~g, self._history[:-1], self.colors)

    def _repr_(self):
        r"""
        Return a string representation of ``self``.

        EXAMPLES::

            sage: RubiksCube().move("R*U")
                         ┌──────────────┐
                         │  3    5   38 │
                         │  2   top  36 │
                         │  1    4   25 │
            ┌────────────┼──────────────┼─────────────┬────────────┐
            │ 33  34  35 │  9   10    8 │ 19   29  32 │ 48  26  27 │
            │ 12 left 13 │ 20  front  7 │ 18 right 31 │ 45 rear 37 │
            │ 14  15  16 │ 22   23    6 │ 17   28  30 │ 43  39  40 │
            └────────────┼──────────────┼─────────────┴────────────┘
                         │ 41   42   11 │
                         │ 44 bottom 21 │
                         │ 46   47   24 │
                         └──────────────┘
        """
        return self._group.repr2d(self._state)

    def facets(self):
        r"""
        Return the facets of ``self``.

        EXAMPLES::

            sage: C = RubiksCube("R*U")
            sage: C.facets()
            [3, 5, 38, 2, 36, 1, 4, 25, 33, 34, 35, 12, 13, 14, 15, 16, 9, 10,
             8, 20, 7, 22, 23, 6, 19, 29, 32, 18, 31, 17, 28, 30, 48, 26, 27,
             45, 37, 43, 39, 40, 41, 42, 11, 44, 21, 46, 47, 24]
        """
        return self._group.facets(self._state)

    def plot(self):
        r"""
        Return a plot of ``self``.

        EXAMPLES::

            sage: C = RubiksCube("R*U")
            sage: C.plot()                                                              # needs sage.plot
            Graphics object consisting of 55 graphics primitives
        """
        return self._group.plot_cube(self._state)

    def show(self):
        r"""
        Show a plot of ``self``.

        EXAMPLES::

            sage: C = RubiksCube("R*U")
            sage: C.show()                                                              # needs sage.plot
        """
        self.plot().show()

    def cubie(self, size, gap, x, y, z, colors, stickers=True):
        """
        Return the cubie at `(x,y,z)`.

        INPUT:

        - ``size`` -- the size of the cubie
        - ``gap`` -- the gap between cubies
        - ``x``, ``y``, ``z`` -- the position of the cubie
        - ``colors`` -- the list of colors
        - ``stickers`` -- boolean (default: ``True``); whether to display stickers

        EXAMPLES::

            sage: C = RubiksCube("R*U")
            sage: C.cubie(0.15, 0.025, 0,0,0, C.colors*3)                               # needs sage.plot
            Graphics3d Object
        """
        sides = cubie_face_list[x, y, z]
        t = 2 * size + gap
        my_colors = [colors[sides[i]+6] for i in range(6)]
        if stickers:
            B = Box(size, size, size, color=(.1, .1, .1))
            S = B + B.stickers(my_colors, size*.1, size*.01)
            return S.translate(-t*x, -t*z, -t*y)
        else:
            return ColorCube(size, [colors[sides[i]+6] for i in range(6)]).translate(-t*x, -t*z, -t*y)

    def plot3d(self, stickers=True):
        r"""
        Return a 3D plot of ``self``.

        EXAMPLES::

            sage: C = RubiksCube("R*U")
            sage: C.plot3d()                                                            # needs sage.plot
            Graphics3d Object
        """
        while len(self.colors) < 7:
            self.colors.append((.1, .1, .1))
        side_colors = [Texture(color=c, ambient=.75) for c in self.colors]
        start_colors = sum([[c]*8 for c in side_colors], [])
        facets = self._group.facets(self._state)
        facet_colors = [0] * 48
        for i in range(48):
            facet_colors[facets[i]-1] = start_colors[i]
        all_colors = side_colors + facet_colors
        pm = [-1,0,1]
        C = sum([self.cubie(.15, .025, x, y, z, all_colors, stickers) for x in pm for y in pm for z in pm], Box(.35, .35, .35, color=self.colors[-1]))
        return C.rotateZ(1.5) #.scale([1,-1,1]).rotateZ(1.5)

    def show3d(self):
        r"""
        Show a 3D plot of ``self``.

        EXAMPLES::

            sage: C = RubiksCube("R*U")
            sage: C.show3d()                                                            # needs sage.plot
        """
        return self.plot3d().show()

    def __richcmp__(self, other, op):
        """
        Comparison.

        INPUT:

        - ``other`` -- anything

        - ``op`` -- comparison operator

        EXAMPLES::

            sage: C = RubiksCube()
            sage: D = RubiksCube("R*U")
            sage: C < D
            True
            sage: C > D
            False
            sage: C == C
            True
            sage: C != D
            True
        """
        if not isinstance(other, RubiksCube):
            return NotImplemented
        return richcmp(self._state, other._state, op)

    def solve(self, algorithm='default', timeout=15):
        r"""
        Solve the Rubik's cube.

        INPUT:

        - ``algorithm`` -- must be one of the following:

          - ``hybrid`` -- (default) try ``kociemba`` for timeout seconds, then ``dietz``
          - ``kociemba`` -- use Dik T. Winter's program (reasonable speed, few moves)
          - ``dietz`` -- use Eric Dietz's cubex program (fast but lots of moves)
          - ``optimal`` -- use Michael Reid's optimal program (may take a long time)
          - ``gap`` -- use GAP word solution (can be slow)

        Any choice other than ``gap`` requires the optional package ``rubiks``.
        If the package is not installed, the ``gap`` algorithm is used by default.

        EXAMPLES::

            sage: C = RubiksCube("R U F L B D")
            sage: C.solve()           # optional - rubiks
            'R U F L B D'

        Dietz's program is much faster, but may give highly non-optimal
        solutions::

            sage: s = C.solve('dietz'); s   # optional - rubiks
            "U' L' L' U L U' L U D L L D' L' D L' D' L D L' U' L D' L' U L' B'
             U' L' U B L D L D' U' L' U L B L B' L' U L U' L' F' L' F L' F L F'
             L' D' L' D D L D' B L B' L B' L B F' L F F B' L F' B D' D' L D B'
             B' L' D' B U' U' L' B' D' F' F' L D F'"
            sage: C2 = RubiksCube(s)  # optional - rubiks
            sage: C == C2             # optional - rubiks
            True
        """
        from sage.features.rubiks import Rubiks
        if Rubiks().is_present():
            import sage.interfaces.rubik  # here to avoid circular referencing
            if algorithm == 'default':
                algorithm = "hybrid"
        else:
            if algorithm == 'default':
                algorithm = 'gap'

        if algorithm == "hybrid":
            try:
                solver = sage.interfaces.rubik.DikSolver()
                return solver.solve(self.facets(), timeout=timeout)
            except RuntimeError:
                solver = sage.interfaces.rubik.CubexSolver()
                return solver.solve(self.facets())

        elif algorithm == "kociemba":
            solver = sage.interfaces.rubik.DikSolver()
            return solver.solve(self.facets(), timeout=timeout)

        elif algorithm == "dietz":
            solver = sage.interfaces.rubik.CubexSolver()
            return solver.solve(self.facets())

        elif algorithm == "optimal":
            # TODO: cache this, startup is expensive
            solver = sage.interfaces.rubik.OptimalSolver()
            return solver.solve(self.facets())

        elif algorithm == "gap":
            solver = CubeGroup()
            return solver.solve(self._state, algorithm='gap')

        else:
            raise ValueError(f"Unrecognized algorithm: {algorithm}")

    def scramble(self, moves=30):
        """
        Scramble the Rubik's cube.

        EXAMPLES::

            sage: C = RubiksCube()
            sage: C.scramble() # random
                         ┌──────────────┐
                         │  3    5   38 │
                         │  2   top  36 │
                         │  1    4   25 │
            ┌────────────┼──────────────┼─────────────┬────────────┐
            │ 33  34  35 │  9   10    8 │ 19   29  32 │ 48  26  27 │
            │ 12 left 13 │ 20  front  7 │ 18 right 31 │ 45 rear 37 │
            │ 14  15  16 │ 22   23    6 │ 17   28  30 │ 43  39  40 │
            └────────────┼──────────────┼─────────────┴────────────┘
                         │ 41   42   11 │
                         │ 44 bottom 21 │
                         │ 46   47   24 │
                         └──────────────┘
        """
        last_move = move = "  "
        all = []
        for i in range(moves):
            while move[0] == last_move[0]:
                move = "RLUDBF"[randint(0, 5)] + " '2"[randint(0, 2)]
            last_move = move
            all.append(move)
        return self.move(' '.join(all))



# sage_setup: distribution = sagemath-modules
# sage.doctest: optional - numpy
r"""
Dense complex double vectors using a NumPy backend

EXAMPLES::

    sage: # needs sage.symbolic
    sage: v = vector(CDF, [(1,-1), (2,pi), (3,5)]); v
    (1.0 - 1.0*I, 2.0 + 3.141592653589793*I, 3.0 + 5.0*I)
    sage: type(v)
    <class 'sage.modules.vector_complex_double_dense.Vector_complex_double_dense'>
    sage: parent(v)
    Vector space of dimension 3 over Complex Double Field
    sage: v[0] = 5
    sage: v
    (5.0, 2.0 + 3.141592653589793*I, 3.0 + 5.0*I)
    sage: loads(dumps(v)) == v
    True

TESTS::

    sage: v = vector(CDF, [2, 2])
    sage: v - v
    (0.0, 0.0)
    sage: (v - v).norm()
    0.0

AUTHORS:

    -- Jason Grout, Oct 2008: switch to NumPy backend, factored out
       Vector_double_dense class
"""

# ***************************************************************************
#       Copyright (C) 2008 Jason Grout <jason-sage@creativetrax.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ***************************************************************************

from sage.rings.complex_double import CDF

cimport numpy


cdef class Vector_complex_double_dense(Vector_double_dense):
    """
    Vectors over the Complex Double Field.  These are supposed to be
    fast vector operations using C doubles. Most operations are
    implemented using numpy which will call the underlying BLAS, if
    needed, on the system.

    EXAMPLES::

        sage: v = vector(CDF, [(1,-1), (2,pi), (3,5)]); v                               # needs sage.symbolic
        (1.0 - 1.0*I, 2.0 + 3.141592653589793*I, 3.0 + 5.0*I)
        sage: v*v  # rel tol 1e-15                                                      # needs sage.symbolic
        -21.86960440108936 + 40.56637061435917*I
    """
    def __cinit__(self, parent, entries, coerce=True, copy=True):
        self._numpy_dtype = numpy.dtype('complex128')
        self._python_dtype = complex
        self._numpy_dtypeint = numpy.NPY_CDOUBLE
        # TODO: Make ComplexDoubleElement instead of CDF for speed
        self._sage_dtype = CDF
        self.__create_vector__()
        return

    def __reduce__(self):
        """
        Pickling.

        EXAMPLES::

            sage: a = vector(CDF, range(9))
            sage: loads(dumps(a)) == a
            True
        """
        return (unpickle_v1, (self._parent, self.list(), self._degree,
                              not self._is_immutable))


# For backwards compatibility, we must keep the function unpickle_v0
def unpickle_v0(parent, entries, degree):
    """
    Create a complex double vector containing the entries.

    EXAMPLES::

        sage: v = vector(CDF, [1,2,3])
        sage: w = sage.modules.vector_complex_double_dense.unpickle_v0(v.parent(), list(v), v.degree())
        sage: v == w
        True
    """
    return unpickle_v1(parent, entries, degree)


def unpickle_v1(parent, entries, degree, is_mutable=None):
    """
    Create a complex double vector with the given parent, entries,
    degree, and mutability.

    EXAMPLES::

        sage: v = vector(CDF, [1,2,3])
        sage: w = sage.modules.vector_complex_double_dense.unpickle_v1(v.parent(), list(v), v.degree(), v.is_immutable())
        sage: v == w
        True
    """
    cdef Vector_complex_double_dense v = Vector_complex_double_dense(parent, entries)
    if is_mutable is not None:
        v._is_immutable = not is_mutable
    return v



# sage_setup: distribution = sagemath-modules
# sage.doctest: needs sage.combinat sage.graphs
r"""
Subalgebras and ideals of Lie algebras

AUTHORS:

- Eero Hakavuori (2018-08-29): initial version
- Travis Scrimshaw (2025-05-21): make all Lie subalgebras use elements
  in the ambient Lie algebra
"""
# ****************************************************************************
#       Copyright (C) 2018 Eero Hakavuori <eero.hakavuori@gmail.com>
#                     2025 Travis Scrimshaw <tcscrims at gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                 https://www.gnu.org/licenses/
# ****************************************************************************
from sage.algebras.lie_algebras.lie_algebra_element import LieSubalgebraElementWrapper
from sage.categories.lie_algebras import LieAlgebras
from sage.categories.homset import Hom
from sage.categories.morphism import SetMorphism
from sage.matrix.constructor import matrix
from sage.misc.cachefunc import cached_method
from sage.misc.lazy_attribute import lazy_attribute
from sage.sets.family import Family
from sage.sets.finite_enumerated_set import FiniteEnumeratedSet
from sage.structure.parent import Parent
from sage.structure.element import parent, Element
from sage.structure.unique_representation import UniqueRepresentation


class LieSubalgebra_finite_dimensional_with_basis(Parent, UniqueRepresentation):
    r"""
    A Lie subalgebra of a finite dimensional Lie algebra with basis.

    INPUT:

    - ``ambient`` -- the Lie algebra containing the subalgebra
    - ``gens`` -- list of generators of the subalgebra
    - ``ideal`` -- boolean (default: ``False``); if ``True``, then ``gens``
      is interpreted as the generating set of an ideal instead of a subalgebra
    - ``order`` -- (optional) the key used to sort the indices of ``ambient``
    - ``category`` -- (optional) a subcategory of subobjects of finite
      dimensional Lie algebras with basis

    EXAMPLES:

    Subalgebras and ideals are defined by giving a list of generators::

        sage: L = lie_algebras.Heisenberg(QQ, 1)
        sage: X, Y, Z = L.basis()
        sage: S =  L.subalgebra([X, Z]); S
        Subalgebra generated by (p1, z) of Heisenberg algebra of rank 1 over Rational Field
        sage: I =  L.ideal([X, Z]); I
        Ideal (p1, z) of Heisenberg algebra of rank 1 over Rational Field

    An ideal is in general larger than the subalgebra with the same generators::

        sage: S = L.subalgebra(Y)
        sage: S.basis()
        Finite family {'q1': q1}
        sage: I = L.ideal(Y)
        sage: I.basis()
        Finite family {'q1': q1, 'z': z}

    The zero dimensional subalgebra can be created by giving 0 as a generator
    or with an empty list of generators::

        sage: L.<X,Y,Z> = LieAlgebra(QQ, {('X','Y'): {'Z': 1}})
        sage: S1 = L.subalgebra(0)
        sage: S2 = L.subalgebra([])
        sage: S1 is S2
        True
        sage: S1.basis()
        Finite family {}

    Elements of the ambient Lie algebra can be reduced modulo an
    ideal or subalgebra::

        sage: # needs sage.symbolic
        sage: L.<X,Y,Z> = LieAlgebra(SR, {('X','Y'): {'Z': 1}})
        sage: I = L.ideal(Y)
        sage: I.reduce(X + 2*Y + 3*Z)
        X
        sage: S = L.subalgebra(Y)
        sage: S.reduce(X + 2*Y + 3*Z)
        X + 3*Z

    The reduction gives elements in a fixed complementary subspace.
    When the base ring is a field, the complementary subspace is spanned by
    those basis elements which are not leading supports of the basis::

        sage: # needs sage.symbolic
        sage: I =  L.ideal(X + Y)
        sage: I.basis()
        Finite family {'Y': X + Y, 'Z': Z}
        sage: el = var('x')*X + var('y')*Y + var('z')*Z; el
        x*X + y*Y + z*Z
        sage: I.reduce(el)
        (x-y)*X

    Giving a different ``order`` may change the reduction of elements::

        sage: I =  L.ideal(X + Y, order=lambda s: ['Z','Y','X'].index(s))               # needs sage.symbolic
        sage: I.basis()                                                                 # needs sage.symbolic
        Finite family {'Z': Z, 'X': X + Y}
        sage: I.reduce(el)                                                              # needs sage.symbolic
        (-x+y)*Y

    A subalgebra of a subalgebra is a subalgebra of the original::

        sage: sc = {('X','Y'): {'Z': 1}, ('X','Z'): {'W': 1}}
        sage: L.<X,Y,Z,W> = LieAlgebra(QQ, sc)
        sage: S1 = L.subalgebra([Y, Z, W]); S1
        Subalgebra generated by (Y, Z, W) of Lie algebra on 4 generators (X, Y, Z, W) over Rational Field
        sage: S2 = S1.subalgebra(S1.gens()[1:]); S2
        Subalgebra generated by (Z, W) of Lie algebra on 4 generators (X, Y, Z, W) over Rational Field
        sage: S3 = S2.subalgebra(S2.gens()[1:]); S3
        Subalgebra generated by (W) of Lie algebra on 4 generators (X, Y, Z, W) over Rational Field

    An ideal of an ideal is not necessarily an ideal of the original::

        sage: I = L.ideal(Y); I
        Ideal (Y) of Lie algebra on 4 generators (X, Y, Z, W) over Rational Field
        sage: J = I.ideal(Z); J
        Ideal (Z) of Ideal (Y) of Lie algebra on 4 generators (X, Y, Z, W) over Rational Field
        sage: J.basis()
        Finite family {'Z': Z}
        sage: J.is_ideal(L)
        False
        sage: K = L.ideal(J.basis().list())
        sage: K.basis()
        Finite family {'Z': Z, 'W': W}

    TESTS:

    Test suites::

        sage: sc = {('X','Y'): {'Z': 1}, ('X','Z'): {'W': 1}}
        sage: L.<X,Y,Z,W> = LieAlgebra(QQ, sc)
        sage: S = L.subalgebra(X + Y)
        sage: TestSuite(S).run()
        sage: I = L.ideal(X + Y)
        sage: TestSuite(I).run()

    Verify that subalgebras and ideals of nilpotent Lie algebras are nilpotent::

        sage: L = LieAlgebra(QQ, 3, step=4)
        sage: x,y,z = L.homogeneous_component_basis(1)
        sage: S = L.subalgebra([x, y])
        sage: S in LieAlgebras(QQ).Nilpotent()
        True
        sage: S.step()
        4
        sage: I = L.ideal(z)
        sage: I in LieAlgebras(QQ).Nilpotent()
        True
        sage: I.step()
        3

    Test computation for a nested ideal::

        sage: sc = {('X','Y'): {'Z': 1}, ('X','Z'): {'W': 1}}
        sage: L.<X,Y,Z,W> = LieAlgebra(QQ, sc)
        sage: I = L.ideal(Y)
        sage: J = I.ideal(Z)
        sage: J.reduce(I(Z) + I(W))
        W
    """
    @staticmethod
    def __classcall_private__(cls, ambient, gens, ideal_of=None,
                              order=None, category=None):
        """
        Normalize input to ensure a unique representation.

        EXAMPLES:

        Various ways to input one generator::

            sage: L.<X,Y> = LieAlgebra(QQ, {('X','Y'): {'X': 1}})
            sage: S1 = L.subalgebra(X)
            sage: S2 = L.subalgebra((X,))
            sage: S3 = L.subalgebra([X])
            sage: S1 is S2 and S2 is S3
            True

        Zero generators are ignored::

            sage: S1 = L.subalgebra(X)
            sage: S2 = L.subalgebra((X, 0))
            sage: S3 = L.subalgebra([X, 0, 0])
            sage: S1 is S2 and S2 is S3
            True
            sage: T1 = L.subalgebra(0)
            sage: T2 = L.subalgebra([])
            sage: T3 = L.subalgebra([0, 0])
            sage: T1 is T2 and T2 is T3
            True

        Subalgebras generated by a subalgebra returns the same subalgebra::

            sage: S1 = L.subalgebra(X)
            sage: S2 = L.subalgebra(S1)
            sage: S1 is S2
            True

        We can have subalgebras generated by subalgebras::

            sage: scoeffs = {('a','d'): {'a':1}, ('a','e'): {'b':-1},
            ....:            ('b','d'): {'b':1}, ('b','e'): {'a':1},
            ....:            ('d','e'): {'c':1}}
            sage: L.<a,b,c,d,e> = LieAlgebra(QQ, scoeffs)
            sage: S1 = L.subalgebra(a)
            sage: S2 = L.subalgebra([b, e])
            sage: S = L.subalgebra([S1, S2])
            sage: S
            Subalgebra generated by (a, a, b, e) of Lie algebra on
             5 generators (a, b, c, d, e) over Rational Field
            sage: S.basis()
            Finite family {'a': a, 'b': b, 'e': e}

        Check that other container-like objects are handled properly
        (:issue:`40137`)::

            sage: L.<a,b,c,d> = LieAlgebra(QQ, {('a','b'): {'c': 1, 'd':1}})
            sage: A = L.ideal([b, c, d])
            sage: B = L.ideal([b, c+d])
            sage: A.ideal(B)
            Ideal (b, c + d) of Ideal (b, c, d) of Lie algebra on 4 generators
             (a, b, c, d) over Rational Field
            sage: A.ideal(B.basis())
            Ideal (b, c + d) of Ideal (b, c, d) of Lie algebra on 4 generators
             (a, b, c, d) over Rational Field
        """
        if isinstance(ambient, LieSubalgebra_finite_dimensional_with_basis):
            ambient = ambient._ambient
        if isinstance(gens, LieSubalgebra_finite_dimensional_with_basis):
            if gens._ideal_of is ideal_of:
                return gens
            gens = gens.lie_algebra_generators()
        if isinstance(gens, Element):
            gens = [gens]
        new_gens = []

        # make sure all elements belong to the ambient Lie algebra
        for gen in gens:
            if isinstance(gen, LieSubalgebra_finite_dimensional_with_basis):
                new_gens.extend(ambient(b) for b in gen.basis())
            elif not gen.is_zero():
                new_gens.append(ambient(gen))
        gens = tuple(new_gens)

        cat = LieAlgebras(ambient.base_ring()).FiniteDimensional().WithBasis()
        category = cat.Subobjects().or_subcategory(category)
        if ambient in LieAlgebras(ambient.base_ring()).Nilpotent():
            category = category.Nilpotent()

        return super().__classcall__(cls, ambient, gens, ideal_of,
                                     order, category)

    def __init__(self, ambient, gens, ideal_of,
                 order=None, category=None) -> None:
        r"""
        Initialize ``self``.

        TESTS::

            sage: L.<X,Y,Z> = LieAlgebra(QQ, {('X','Y'): {'Z': 1}})
            sage: S = L.subalgebra(X)
            sage: TestSuite(S).run()
            sage: I = L.ideal(X)
            sage: TestSuite(I).run()

        Check that :issue:`34006` is fixed::

            sage: S.gens()[0].parent() is S
            True
        """
        self._ambient = ambient
        self._ideal_of = ideal_of

        # initialize helper variables for ordering
        if order is None:
            if hasattr(ambient, "_basis_key"):
                order = ambient._basis_key
            else:
                order = lambda x: x
        self._order = order
        self._reversed_indices = sorted(ambient.indices(), key=order,
                                        reverse=True)
        # helper to reorder a vector that has been jumbled by the above
        self._reorganized_indices = [self._reversed_indices.index(i)
                                     for i in ambient.indices()]

        super().__init__(ambient.base_ring(), category=category)

        self._gens = tuple([self.element_class(self, g) for g in gens])

        # register a coercion to the ambient Lie algebra
        H = Hom(self, ambient)
        f = SetMorphism(H, self.lift)
        ambient.register_coercion(f)

    def __contains__(self, x) -> bool:
        r"""
        Return ``True`` if ``x`` is an element of ``self``.

        EXAMPLES:

        Elements of the ambient Lie algebra are contained in the subalgebra
        if they are iterated brackets of the generators::

            sage: sc = {('x','y'): {'z': 1}, ('x','z'): {'w': 1}}
            sage: L.<x,y,z,w,u> = LieAlgebra(QQ, sc)
            sage: S = L.subalgebra([x, y])
            sage: z in S
            True
            sage: w in S
            True
            sage: u in S
            False

        TESTS::

            sage: L.<x,y,z> = LieAlgebra(QQ, {('x','y'): {'z': 1}})
            sage: I = L.subalgebra(x)
            sage: I(x) in I
            True
        """
        if x in self._ambient:
            x = self._ambient(x)
            return x.to_vector() in self.module()
        return super().__contains__(x)

    def __getitem__(self, x):
        r"""
        If `x` is a pair `(a, b)`, return the Lie bracket `[a, b]`.
        Otherwise try to return the `x`-th element of ``self``.

        This replicates the convenience syntax for Lie brackets
        of Lie algebras.

        EXAMPLES::

            sage: L.<x,y, z> = LieAlgebra(QQ, {('x','y'): {'z': 1}})
            sage: S = L.subalgebra([x, y])
            sage: a = S(x); b = S(y)
            sage: S[a, b]
            z
            sage: S[a, a + S[a,b]]
            0
        """
        if isinstance(x, tuple) and len(x) == 2:
            return self(x[0])._bracket_(self(x[1]))
        return super().__getitem__(x)

    def _repr_(self):
        r"""
        Return a string representation of ``self``.

        EXAMPLES::

            sage: L.<X,Y> = LieAlgebra(QQ, abelian=True)
            sage: L.subalgebra([X, Y])
            Subalgebra generated by (X, Y) of Abelian Lie algebra on 2 generators (X, Y) over Rational Field
            sage: L.ideal([X, Y])
            Ideal (X, Y) of Abelian Lie algebra on 2 generators (X, Y) over Rational Field
        """
        if self._ideal_of is not None:
            return "Ideal {} of {}".format(self._repr_short(), self._ideal_of)

        return "Subalgebra generated by {} of {}".format(self._repr_short(), self.ambient())

    def _repr_short(self):
        """
        Represent the list of generators.

        EXAMPLES::

            sage: L.<X,Y> = LieAlgebra(QQ, abelian=True)
            sage: L.ideal([X, Y])._repr_short()
            '(X, Y)'
            sage: L.ideal(X)._repr_short()
            '(X)'
            sage: L.subalgebra(X)._repr_short()
            '(X)'
        """
        return '(%s)' % (', '.join(str(X) for X in self.gens()))

    def _an_element_(self):
        r"""
        Return an element of ``self``.

        EXAMPLES::

            sage: L.<X,Y> = LieAlgebra(QQ, abelian=True)
            sage: S = L.subalgebra([X, Y])
            sage: S._an_element_()
            X
        """
        return next(iter(self.lie_algebra_generators()))

    def _element_constructor_(self, x):
        """
        Convert ``x`` into ``self``.

        EXAMPLES:

        Elements of subalgebras are created directly from elements
        of the ambient Lie algebra::

            sage: L.<x,y,z,w> = LieAlgebra(ZZ, {('x','y'): {'z': 1}})
            sage: S = L.subalgebra([x, y])
            sage: S(y)
            y
            sage: S(y).parent()
            Subalgebra generated by (x, y) of Lie algebra on 4 generators (x, y, z, w) over Integer Ring

        A vector contained in the module corresponding to the subalgebra is
        interpreted as a coordinate vector::

            sage: S.module()
            Free module of degree 4 and rank 3 over Integer Ring
            User basis matrix:
            [1 0 0 0]
            [0 1 0 0]
            [0 0 1 0]
            sage: S(vector(ZZ, [2, 3, 5, 0]))
            2*x + 3*y + 5*z

        A list of 2 elements is interpreted as a Lie bracket::

            sage: S([S(x), S(y)])
            z
            sage: S([S(x), S(y)]) == S(L[x, y])
            True
        """
        P = parent(x)
        if P is self:
            return x
        if P == self._ambient:
            return self.retract(x)
        if isinstance(P, type(self)) and P._ambient == self._ambient:
            return self.retract(x.value)

        if x in self.module():
            return self.from_vector(x)

        if isinstance(x, list) and len(x) == 2:
            return self(x[0])._bracket_(self(x[1]))

        return super()._element_constructor_(x)

    def _to_m(self, X):
        r"""
        Return the vector of an element of the ambient Lie algebra with indices
        reorganized in decreasing order according to ``self._order``.

        This is used internally for submodule computations, so that the pivot
        elements in the echelon form are given by
        :meth:`~sage.categories.modules_with_basis.ModulesWithBasis.ElementMethods.leading_term`
        using the sort key ``self._order``.

        INPUT:

        - ``X`` -- an element of the ambient Lie algebra

        EXAMPLES:

        If the basis is in increasing order, it is reversed::

            sage: L.<x,y,z> = LieAlgebra(QQ, {('x','y'): {'z': 1}})
            sage: I = L.ideal([x, z])
            sage: el = x + 2*y + 3*z
            sage: el.to_vector()
            (1, 2, 3)
            sage: I._to_m(el)
            (3, 2, 1)

        This follows the reverse order of the ambient basis order::

            sage: L.<x,z,y> = LieAlgebra(QQ, {('x','y'): {'z': 1}})
            sage: I = L.ideal([x, z])
            sage: el = x + 2*y + 3*z
            sage: el.to_vector()
            (1, 3, 2)
            sage: I._to_m(el)
            (2, 3, 1)
        """
        mc = X.monomial_coefficients()
        M = self._ambient.module()
        R = M.base_ring()
        B = M.basis()
        return M.sum(R(mc[self._reversed_indices[i]]) * B[i]
                     for i in range(len(B)) if self._reversed_indices[i] in mc)

    def _from_m(self, v):
        r"""
        Return the element of the ambient Lie algebra from a reorganized vector.

        This is used internally for submodule computations, so that the pivot
        elements in the echelon form are given by
        :meth:`~sage.categories.modules_with_basis.ModulesWithBasis.ElementMethods.leading_term`
        using the sort key `self._order`.

        INPUT:

        - ``v`` -- a vector

        EXAMPLES:

        Unscrambling of a vector when the basis is in increasing order::

            sage: L.<x,y,z> = LieAlgebra(QQ, {('x','y'): {'z': 1}})
            sage: I = L.ideal([x, z])
            sage: I._from_m([3, 2, 1])
            x + 2*y + 3*z

        The map is the inverse of :meth:`_to_m`::

            sage: L.<c,a,e,f,b,d> = LieAlgebra(QQ, abelian=True)
            sage: S = L.subalgebra(L.basis().list())
            sage: v = S._to_m(c + 2*a + 3*e + 4*f + 5*b + 6*d); v
            (6, 5, 4, 3, 2, 1)
            sage: S._from_m(v)
            c + 2*a + 3*e + 4*f + 5*b + 6*d
            sage: all(S._from_m(S._to_m(X)) == X for X in L.some_elements())
            True
        """
        L = self.ambient()
        M = L.module()
        R = M.base_ring()
        v_self = M.coordinate_vector(v)
        B = M.basis()
        v_sorted = M.sum(R(v_self[self._reorganized_indices[i]]) * B[i]
                         for i in range(len(B)))
        return L.from_vector(v_sorted)

    @lazy_attribute
    def _indices(self):
        r"""
        Return the set of indices for the basis of ``self``.

        EXAMPLES::

            sage: L.<x,y,z> = LieAlgebra(QQ, abelian=True)
            sage: S = L.subalgebra([x, y])
            sage: S._indices
            {'x', 'y'}
            sage: [S.basis()[k] for k in S._indices]
            [x, y]
        """
        return FiniteEnumeratedSet(self.basis().keys())

    def indices(self):
        r"""
        Return the set of indices for the basis of ``self``.

        EXAMPLES::

            sage: L.<x,y,z> = LieAlgebra(QQ, abelian=True)
            sage: S = L.subalgebra([x+y, z])
            sage: S.indices()
            {'y', 'z'}
            sage: [S.basis()[k] for k in S.indices()]
            [x + y, z]
        """
        return self._indices

    @cached_method
    def zero(self):
        r"""
        Return the element `0`.

        EXAMPLES::

            sage: L.<x,y> = LieAlgebra(QQ, abelian=True)
            sage: S = L.subalgebra(x)
            sage: S.zero()
            0
            sage: S.zero() == S(L.zero())
            True
        """
        return self.element_class(self, self.ambient().zero())

    def ambient(self):
        r"""
        Return the ambient Lie algebra of ``self``.

        EXAMPLES::

            sage: L.<x,y> = LieAlgebra(QQ, abelian=True)
            sage: S = L.subalgebra(x)
            sage: S.ambient() is L
            True
        """
        return self._ambient

    def lift(self, X):
        r"""
        Coerce an element ``X`` of ``self`` into the ambient Lie algebra.

        INPUT:

        - ``X`` -- an element of ``self``

        EXAMPLES::

            sage: L.<x,y> = LieAlgebra(QQ, abelian=True)
            sage: S = L.subalgebra(x)
            sage: sx = S(x); sx
            x
            sage: sx.parent()
            Subalgebra generated by (x) of Abelian Lie algebra on 2 generators (x, y) over Rational Field
            sage: a = S.lift(sx); a
            x
            sage: a.parent()
            Abelian Lie algebra on 2 generators (x, y) over Rational Field
        """
        return X.value

    def retract(self, X):
        r"""
        Retract ``X`` to ``self``.

        INPUT:

        - ``X`` -- an element of the ambient Lie algebra

        EXAMPLES:

        Retraction to a subalgebra of a free nilpotent Lie algebra::

            sage: L = LieAlgebra(QQ, 3, step=2)
            sage: L.inject_variables()
            Defining X_1, X_2, X_3, X_12, X_13, X_23
            sage: S = L.subalgebra([X_1, X_2])
            sage: el = S.retract(2*X_1 + 3*X_2 + 5*X_12); el
            2*X_1 + 3*X_2 + 5*X_12
            sage: el.parent()
            Subalgebra generated by (X_1, X_2) of Free Nilpotent Lie algebra
             of rank 3 and step 2 over Rational Field

        Retraction raises an error if the element is not contained in the
        subalgebra::

            sage: S.retract(X_3)
            Traceback (most recent call last):
            ...
            ValueError: the element X_3 is not in Subalgebra generated
             by (X_1, X_2) of Free Nilpotent Lie algebra of rank 3 and
             step 2 over Rational Field
        """
        X = self._ambient(X)
        if X not in self:
            raise ValueError("the element %s is not in %s" % (X, self))

        return self.element_class(self, X)

    def gens(self) -> tuple:
        r"""
        Return the generating set of ``self``.

        EXAMPLES::

            sage: L.<x,y,z> = LieAlgebra(QQ, {('x','y'): {'z': 1}})
            sage: S = L.subalgebra(x)
            sage: S.gens()
            (x,)
        """
        return self._gens

    def lie_algebra_generators(self):
        r"""
        Return the generating set of ``self`` as a Lie algebra.

        EXAMPLES:

        The Lie algebra generators of a subalgebra are the original generators::

            sage: L.<x,y,z> = LieAlgebra(QQ, {('x','y'): {'z': 1}})
            sage: S = L.subalgebra(x)
            sage: S.lie_algebra_generators()
            (x,)

        The Lie algebra generators of an ideal is usually a larger set::

            sage: I = L.ideal(x)
            sage: I.lie_algebra_generators()
            Finite family {'x': x, 'z': z}
        """
        if self._ideal_of is not None:
            return self.basis()
        return self._gens

    @cached_method
    def basis(self):
        r"""
        Return a basis of ``self``.

        EXAMPLES:

        A basis of a subalgebra::

            sage: sc = {('a','b'): {'c': 1}, ('a','c'): {'d': 1}}
            sage: L.<a,b,c,d> = LieAlgebra(QQ, sc)
            sage: L.subalgebra([a + b, c + d]).basis()
            Finite family {'b': a + b, 'c': c, 'd': d}

        A basis of an ideal::

            sage: sc = {('x','y'): {'z': 1}, ('x','z'): {'w': 1}}
            sage: L.<x,y,z,w> = LieAlgebra(QQ, sc)
            sage: L.ideal([x + y + z + w]).basis()
            Finite family {'y': x + y, 'z': z, 'w': w}

        This also works for Lie algebras whose natural basis elements
        are not comparable (but have a well-defined basis ordering)::

            sage: sl3 = LieAlgebra(QQ, cartan_type=['A',2])
            sage: D = sl3.derived_subalgebra()
            sage: len(D.basis())
            8
            sage: e = list(sl3.e())
            sage: sl3.ideal(e).dimension()
            8
            sage: sl3.subalgebra(e).dimension()
            3
        """
        ambient = self._ambient
        if self._ideal_of is not None:
            L = self._ideal_of
            B = [self._to_m(ambient(X)) for X in L.basis()]
        else:
            L = ambient
            B = [self._to_m(X) for X in ambient.basis()]

        m = ambient.module()
        sm = m.submodule([self._to_m(X.value) for X in self.gens()])
        d = 0

        while sm.dimension() > d:
            d = sm.dimension()
            SB = sm.basis()
            if self._ideal_of is None:
                B = SB

            brackets = [self._to_m(ambient.bracket(self._from_m(v), self._from_m(w)))
                        for v in B for w in SB]
            sm = m.submodule(sm.basis() + brackets)

        basis = [self.element_class(self, self._from_m(v))
                 for v in sm.echelonized_basis()]

        indices = [self.lift(X).leading_support(key=self._order) for X in basis]
        basis = dict(zip(indices, basis))
        indices.sort(key=self._order)

        return Family(indices, basis.__getitem__)

    @cached_method
    def leading_monomials(self):
        r"""
        Return the set of leading monomials of the basis of ``self``.

        EXAMPLES:

        A basis of an ideal and the corresponding leading monomials::

            sage: sc = {('a','b'): {'c': 2}, ('a','c'): {'d': 4}}
            sage: L.<a,b,c,d> = LieAlgebra(ZZ, sc)
            sage: I = L.ideal(a + b)
            sage: I.basis()
            Finite family {'b': a + b, 'c': 2*c, 'd': 4*d}
            sage: I.leading_monomials()
            Family (b, c, d)

        A different ordering can give different leading monomials::

            sage: key = lambda s: ['d','c','b','a'].index(s)
            sage: I = L.ideal(a + b, order=key)
            sage: I.basis()
            Finite family {'d': 4*d, 'c': 2*c, 'a': a + b}
            sage: I.leading_monomials()
            Family (d, c, a)
        """
        return Family(self.lift(X).leading_monomial(key=self._order)
                      for X in self.basis())

    def from_vector(self, v, order=None, coerce=False):
        r"""
        Return the element of ``self`` corresponding to the vector ``v``.

        INPUT:

        - ``v`` -- a vector in ``self.module()`` or ``self.ambient().module()``

        EXAMPLES:

        An element from a vector of the intrinsic module::

            sage: L.<X,Y,Z> = LieAlgebra(ZZ, abelian=True)
            sage: L.dimension()
            3
            sage: S = L.subalgebra([X, Y])
            sage: S.dimension()
            2
            sage: el = S.from_vector([1, 2]); el
            X + 2*Y
            sage: el.parent() == S
            True

        An element from a vector of the ambient module

            sage: el = S.from_vector([1, 2, 0]); el
            X + 2*Y
            sage: el.parent() == S
            True
        """
        if len(v) == self._ambient.dimension():
            return self.retract(self._ambient.from_vector(v))
        return super().from_vector(v)

    def basis_matrix(self):
        r"""
        Return the basis matrix of ``self`` as a submodule
        of the ambient Lie algebra.

        EXAMPLES::

            sage: L.<X,Y,Z> = LieAlgebra(ZZ, {('X','Y'): {'Z': 3}})
            sage: S1 = L.subalgebra([4*X + Y, Y])
            sage: S1.basis_matrix()
            [ 4  0  0]
            [ 0  1  0]
            [ 0  0 12]
            sage: K.<X,Y,Z> = LieAlgebra(QQ, {('X','Y'): {'Z': 3}})
            sage: S2 = K.subalgebra([4*X + Y, Y])
            sage: S2.basis_matrix()
            [1 0 0]
            [0 1 0]
            [0 0 1]
        """
        return self.module().basis_matrix()

    @cached_method
    def module(self, sparse=False):
        r"""
        Return the submodule of the ambient Lie algebra
        corresponding to ``self``.

        EXAMPLES::

            sage: L.<X,Y,Z> = LieAlgebra(ZZ, {('X','Y'): {'Z': 3}})
            sage: S = L.subalgebra([X, Y])
            sage: S.module()
            Free module of degree 3 and rank 3 over Integer Ring
            User basis matrix:
            [1 0 0]
            [0 1 0]
            [0 0 3]
        """
        try:
            m = self._ambient.module(sparse=sparse)
        except TypeError:
            m = self._ambient.module()
        ambientbasis = [self.lift(X).to_vector() for X in self.basis()]
        return m.submodule_with_basis(ambientbasis)

    @cached_method
    def is_ideal(self, A):
        """
        Return if ``self`` is an ideal of ``A``.

        EXAMPLES:

        Some subalgebras are ideals::

            sage: L.<x,y,z> = LieAlgebra(QQ, {('x','y'): {'z': 1}})
            sage: S1 = L.subalgebra([x])
            sage: S1.is_ideal(L)
            False
            sage: S2 = L.subalgebra([x, y])
            sage: S2.is_ideal(L)
            True
            sage: S3 = L.subalgebra([y, z])
            sage: S3.is_ideal(L)
            True

        All ideals are ideals::

            sage: L.<x,y> = LieAlgebra(QQ, {('x','y'): {'x': 1}})
            sage: I = L.ideal(x)
            sage: I.is_ideal(L)
            True
            sage: I.is_ideal(I)
            True

        TESTS::

            sage: L.<x,y> = LieAlgebra(QQ, {('x','y'): {'x': 1}})
            sage: I = L.ideal(x)
            sage: L.is_ideal(I)
            False
        """
        if A is self._ideal_of:
            return True
        return super().is_ideal(A)

    class Element(LieSubalgebraElementWrapper):
        def adjoint_matrix(self, sparse=False):
            """
            Return the matrix of the adjoint action of ``self``.

            EXAMPLES::

                sage: MS = MatrixSpace(QQ, 2)
                sage: m = MS([[0, -1], [1, 0]])
                sage: L = LieAlgebra(associative=MS)
                sage: S = L.subalgebra([m])
                sage: S.basis()
                Finite family {(1, 0): [ 0 -1]
                                       [ 1  0]}
                sage: x = S.basis()[1,0]
                sage: x.parent() is S
                True
                sage: x.adjoint_matrix()
                [0]

                sage: m1 = MS([[0, 1], [0, 0]])
                sage: m2 = MS([[0, 0], [1, 0]])
                sage: S = L.subalgebra([m1, m2])
                sage: e,f = S.lie_algebra_generators()
                sage: ascii_art([b.value.value for b in S.basis()])
                [ [0 1]  [0 0]  [-1  0] ]
                [ [0 0], [1 0], [ 0  1] ]
                sage: E = e.adjoint_matrix(); E
                [ 0  0  2]
                [ 0  0  0]
                [ 0 -1  0]
                sage: F = f.adjoint_matrix(); F
                [ 0  0  0]
                [ 0  0 -2]
                [ 1  0  0]
                sage: h = e.bracket(f)
                sage: E * F - F * E == h.adjoint_matrix()
                True

            TESTS:

            Check that :issue:`34006` is fixed::

                sage: MS = MatrixSpace(QQ, 2)
                sage: m = MS([[0, -1], [1, 0]])
                sage: L = LieAlgebra(associative=MS)
                sage: S = L.subalgebra([m])
                sage: S.killing_form_matrix()
                [0]
            """
            P = self.parent()
            basis = P.basis()
            M = P.module(sparse=sparse)
            return matrix(self.base_ring(),
                          [M.coordinate_vector(P.bracket(self, b).to_vector(sparse=sparse))
                           for b in basis], sparse=sparse).transpose()



# sage_setup: distribution = sagemath-schemes
# sage.doctest: needs sage.libs.flint sage.libs.pari
"""
Spaces of homomorphisms between modular abelian varieties

EXAMPLES:

First, we consider J0(37). This Jacobian has two simple factors,
corresponding to distinct newforms. These two intersect
nontrivially in J0(37).

::

    sage: J = J0(37)
    sage: D = J.decomposition() ; D
    [Simple abelian subvariety 37a(1,37) of dimension 1 of J0(37),
     Simple abelian subvariety 37b(1,37) of dimension 1 of J0(37)]
    sage: D[0].intersection(D[1])
    (Finite subgroup with invariants [2, 2] over QQ of
      Simple abelian subvariety 37a(1,37) of dimension 1 of J0(37),
     Simple abelian subvariety of dimension 0 of J0(37))

As an abstract product, since these newforms are distinct, the
corresponding simple abelian varieties are not isogenous, and so
there are no maps between them. The endomorphism ring of the
corresponding product is thus isomorphic to the direct sum of the
endomorphism rings for each factor. Since the factors correspond to
abelian varieties of dimension 1, these endomorphism rings are each
isomorphic to ZZ.

::

    sage: Hom(D[0],D[1]).gens()
    ()
    sage: A = D[0] * D[1] ; A
    Abelian subvariety of dimension 2 of J0(37) x J0(37)
    sage: A.endomorphism_ring().gens()
    (Abelian variety endomorphism of Abelian subvariety of dimension 2 of J0(37) x J0(37),
     Abelian variety endomorphism of Abelian subvariety of dimension 2 of J0(37) x J0(37))
    sage: [ x.matrix() for x in A.endomorphism_ring().gens() ]
    [
    [1 0 0 0]  [0 0 0 0]
    [0 1 0 0]  [0 0 0 0]
    [0 0 0 0]  [0 0 1 0]
    [0 0 0 0], [0 0 0 1]
    ]

However, these two newforms have a congruence between them modulo
2, which gives rise to interesting endomorphisms of J0(37).

::

    sage: E = J.endomorphism_ring()
    sage: E.gens()
    (Abelian variety endomorphism of Abelian variety J0(37) of dimension 2,
     Abelian variety endomorphism of Abelian variety J0(37) of dimension 2)
    sage: [ x.matrix() for x in E.gens() ]
    [
    [1 0 0 0]  [ 0  1  1 -1]
    [0 1 0 0]  [ 1  0  1  0]
    [0 0 1 0]  [ 0  0 -1  1]
    [0 0 0 1], [ 0  0  0  1]
    ]
    sage: (-1*E.gens()[0] + E.gens()[1]).matrix()
    [-1  1  1 -1]
    [ 1 -1  1  0]
    [ 0  0 -2  1]
    [ 0  0  0  0]

Of course, these endomorphisms will be reflected in the Hecke
algebra, which is in fact the full endomorphism ring of J0(37) in
this case::

    sage: J.hecke_operator(2).matrix()
    [-1  1  1 -1]
    [ 1 -1  1  0]
    [ 0  0 -2  1]
    [ 0  0  0  0]
    sage: T = E.image_of_hecke_algebra()
    sage: T.gens()
    (Abelian variety endomorphism of Abelian variety J0(37) of dimension 2,
     Abelian variety endomorphism of Abelian variety J0(37) of dimension 2)
    sage: [ x.matrix() for x in T.gens() ]
    [
    [1 0 0 0]  [ 0  1  1 -1]
    [0 1 0 0]  [ 1  0  1  0]
    [0 0 1 0]  [ 0  0 -1  1]
    [0 0 0 1], [ 0  0  0  1]
    ]
    sage: T.index_in(E)
    1

Next, we consider J0(33). In this case, we have both oldforms and
newforms. There are two copies of J0(11), one for each degeneracy
map from J0(11) to J0(33). There is also one newform at level 33.
The images of the two degeneracy maps are, of course, isogenous.

::

    sage: J = J0(33)
    sage: D = J.decomposition()
    sage: D
    [Simple abelian subvariety 11a(1,33) of dimension 1 of J0(33),
     Simple abelian subvariety 11a(3,33) of dimension 1 of J0(33),
     Simple abelian subvariety 33a(1,33) of dimension 1 of J0(33)]
    sage: Hom(D[0],D[1]).gens()
    (Abelian variety morphism:
      From: Simple abelian subvariety 11a(1,33) of dimension 1 of J0(33)
      To:   Simple abelian subvariety 11a(3,33) of dimension 1 of J0(33),)
    sage: Hom(D[0],D[1]).gens()[0].matrix()
    [ 0  1]
    [-1  0]

Then this gives that the component corresponding to the sum of the
oldforms will have a rank 4 endomorphism ring. We also have a rank
one endomorphism ring for the newform 33a (since it is again
1-dimensional), which gives a rank 5 endomorphism ring for J0(33).

::

    sage: DD = J.decomposition(simple=False) ; DD
    [Abelian subvariety of dimension 2 of J0(33),
     Abelian subvariety of dimension 1 of J0(33)]
    sage: A, B = DD
    sage: A == D[0] + D[1]
    True
    sage: A.endomorphism_ring().gens()
    (Abelian variety endomorphism of Abelian subvariety of dimension 2 of J0(33),
     Abelian variety endomorphism of Abelian subvariety of dimension 2 of J0(33),
     Abelian variety endomorphism of Abelian subvariety of dimension 2 of J0(33),
     Abelian variety endomorphism of Abelian subvariety of dimension 2 of J0(33))
    sage: B.endomorphism_ring().gens()
    (Abelian variety endomorphism of Abelian subvariety of dimension 1 of J0(33),)
    sage: E = J.endomorphism_ring() ; E.gens()  # long time (3s on sage.math, 2011)
    (Abelian variety endomorphism of Abelian variety J0(33) of dimension 3,
     Abelian variety endomorphism of Abelian variety J0(33) of dimension 3,
     Abelian variety endomorphism of Abelian variety J0(33) of dimension 3,
     Abelian variety endomorphism of Abelian variety J0(33) of dimension 3,
     Abelian variety endomorphism of Abelian variety J0(33) of dimension 3)

In this case, the image of the Hecke algebra will only have rank 3,
so that it is of infinite index in the full endomorphism ring.
However, if we call this image T, we can still ask about the index
of T in its saturation, which is 1 in this case.

::

    sage: # long time
    sage: T = E.image_of_hecke_algebra()
    sage: T.gens()
    (Abelian variety endomorphism of Abelian variety J0(33) of dimension 3,
     Abelian variety endomorphism of Abelian variety J0(33) of dimension 3,
     Abelian variety endomorphism of Abelian variety J0(33) of dimension 3)
    sage: T.index_in(E)
    +Infinity
    sage: T.index_in_saturation()
    1

TESTS::

    sage: J = J0(37) ; J.Hom(J)(matrix(ZZ,4,[5..20]))
    Abelian variety endomorphism of Abelian variety J0(37) of dimension 2
    sage: K = J0(11) * J0(11) ; J.Hom(K)(matrix(ZZ,4,[5..20]))
    Abelian variety morphism:
      From: Abelian variety J0(37) of dimension 2
      To:   Abelian variety J0(11) x J0(11) of dimension 2

AUTHORS:

- William Stein (2007-03)

- Craig Citro, Robert Bradshaw (2008-03): Rewrote with modabvar overhaul
"""

# ****************************************************************************
#       Copyright (C) 2007 William Stein <wstein@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************


from copy import copy

from sage.categories.homset import HomsetWithBase
from sage.structure.element import parent
from sage.structure.parent import Parent
from sage.misc.lazy_attribute import lazy_attribute


from sage.modular.abvar import morphism

from sage.rings.infinity import Infinity

from sage.matrix.matrix_space import MatrixSpace
from sage.matrix.constructor import matrix, identity_matrix
from sage.structure.element import Matrix

from sage.rings.integer_ring import ZZ


class Homspace(HomsetWithBase):
    """
    A space of homomorphisms between two modular abelian varieties.
    """
    Element = morphism.Morphism

    def __init__(self, domain, codomain, cat):
        """
        Create a homspace.

        INPUT:

        - ``domain, codomain`` -- modular abelian varieties

        - ``cat`` -- category

        EXAMPLES::

            sage: H = Hom(J0(11), J0(22)); H
            Space of homomorphisms from Abelian variety J0(11) of dimension 1
             to Abelian variety J0(22) of dimension 2
            sage: Hom(J0(11), J0(11))
            Endomorphism ring of Abelian variety J0(11) of dimension 1
            sage: type(H)
            <class 'sage.modular.abvar.homspace.Homspace_with_category'>
            sage: H.homset_category()
            Category of modular abelian varieties over Rational Field
        """
        from .abvar import ModularAbelianVariety_abstract
        if not isinstance(domain, ModularAbelianVariety_abstract):
            raise TypeError("domain must be a modular abelian variety")
        if not isinstance(codomain, ModularAbelianVariety_abstract):
            raise TypeError("codomain must be a modular abelian variety")
        self._gens = None
        HomsetWithBase.__init__(self, domain, codomain, category=cat)

    def identity(self):
        """
        Return the identity endomorphism.

        EXAMPLES::

            sage: E = End(J0(11))
            sage: E.identity()
            Abelian variety endomorphism of Abelian variety J0(11) of dimension 1
            sage: E.one()
            Abelian variety endomorphism of Abelian variety J0(11) of dimension 1

            sage: H = Hom(J0(11), J0(22))
            sage: H.identity()
            Traceback (most recent call last):
            ...
            TypeError: the identity map is only defined for endomorphisms
        """
        if self.domain() is not self.codomain():
            raise TypeError("the identity map is only defined for endomorphisms")
        M = self.matrix_space().one()
        return self.element_class(self, M)

    @lazy_attribute
    def _matrix_space(self):
        """
        Return the matrix space of ``self``.

        .. WARNING::

            During unpickling, the domain and codomain may be unable to
            provide the necessary information. This is why this is a lazy
            attribute. See :issue:`14793`.

        EXAMPLES::

            sage: Hom(J0(11), J0(22))._matrix_space
            Full MatrixSpace of 2 by 4 dense matrices over Integer Ring
        """
        return MatrixSpace(ZZ, 2*self.domain().dimension(), 2*self.codomain().dimension())

    def _element_constructor_from_element_class(self, *args, **keywords):
        """
        Used in the coercion framework. Unfortunately, the default method
        would get the order of parent and data different from what is expected
        in ``MatrixMorphism.__init__``.

        EXAMPLES::

            sage: H = Hom(J0(11), J0(22))
            sage: phi = H(matrix(ZZ,2,4,[5..12])); phi # indirect doctest
            Abelian variety morphism:
              From: Abelian variety J0(11) of dimension 1
              To:   Abelian variety J0(22) of dimension 2
        """
        return self.element_class(self, *args, **keywords)

    def __call__(self, M, **kwds):
        r"""
        Create a homomorphism in this space from M. M can be any of the
        following:

        - a Morphism of abelian varieties

        - a matrix of the appropriate size
          (i.e. 2\*self.domain().dimension() x
          2\*self.codomain().dimension()) whose entries are coercible
          into self.base_ring()

        - anything that can be coerced into self.matrix_space()

        EXAMPLES::

            sage: H = Hom(J0(11), J0(22))
            sage: phi = H(matrix(ZZ,2,4,[5..12])) ; phi
            Abelian variety morphism:
              From: Abelian variety J0(11) of dimension 1
              To:   Abelian variety J0(22) of dimension 2
            sage: phi.matrix()
            [ 5  6  7  8]
            [ 9 10 11 12]
            sage: phi.matrix().parent()
            Full MatrixSpace of 2 by 4 dense matrices over Integer Ring

        ::

            sage: H = J0(22).Hom(J0(11)*J0(11))
            sage: m1 = J0(22).degeneracy_map(11,1).matrix() ; m1
            [ 0  1]
            [-1  1]
            [-1  0]
            [ 0 -1]
            sage: m2 = J0(22).degeneracy_map(11,2).matrix() ; m2
            [ 1 -2]
            [ 0 -2]
            [ 1 -1]
            [ 0 -1]
            sage: m = m1.transpose().stack(m2.transpose()).transpose() ; m
            [ 0  1  1 -2]
            [-1  1  0 -2]
            [-1  0  1 -1]
            [ 0 -1  0 -1]
            sage: phi = H(m) ; phi
            Abelian variety morphism:
              From: Abelian variety J0(22) of dimension 2
              To:   Abelian variety J0(11) x J0(11) of dimension 2
            sage: phi.matrix()
            [ 0  1  1 -2]
            [-1  1  0 -2]
            [-1  0  1 -1]
            [ 0 -1  0 -1]
        """
        side = kwds.get("side", "left")
        if isinstance(M, morphism.Morphism):
            if M.parent() is self:
                return M
            elif M.domain() == self.domain() and M.codomain() == self.codomain():
                M = M.matrix()
            else:
                raise ValueError("cannot convert %s into %s" % (M, self))
        elif isinstance(M, Matrix):
            if M.base_ring() != ZZ:
                M = M.change_ring(ZZ)
            if side == "left":
                if M.nrows() != 2*self.domain().dimension() or M.ncols() != 2*self.codomain().dimension():
                    raise TypeError("matrix has wrong dimension")
            else:
                if M.ncols() != 2*self.domain().dimension() or M.nrows() != 2*self.codomain().dimension():
                    raise TypeError("matrix has wrong dimension")
        elif self.matrix_space().has_coerce_map_from(parent(M)):
            M = self.matrix_space()(M)
        else:
            raise TypeError("can only coerce in matrices or morphisms")
        return self.element_class(self, M, side)

    def _repr_(self):
        """
        String representation of a modular abelian variety homspace.

        EXAMPLES::

            sage: J = J0(11)
            sage: End(J)._repr_()
            'Endomorphism ring of Abelian variety J0(11) of dimension 1'
        """
        return "Space of homomorphisms from %s to %s" %\
               (self.domain(), self.codomain())

    def _get_matrix(self, g):
        """
        Given an object g, try to return a matrix corresponding to g with
        dimensions the same as those of self.matrix_space().

        INPUT:

        - ``g`` -- a matrix or morphism or object with a list method

        OUTPUT: a matrix

        EXAMPLES::

            sage: E = End(J0(11))
            sage: E._get_matrix(matrix(QQ,2,[1,2,3,4]))
            [1 2]
            [3 4]
            sage: E._get_matrix(J0(11).hecke_operator(2))
            [-2  0]
            [ 0 -2]

        ::

            sage: H = Hom(J0(11) * J0(17), J0(22))
            sage: H._get_matrix(tuple([8..23]))
            [ 8  9 10 11]
            [12 13 14 15]
            [16 17 18 19]
            [20 21 22 23]
            sage: H._get_matrix(tuple([8..23]))
            [ 8  9 10 11]
            [12 13 14 15]
            [16 17 18 19]
            [20 21 22 23]
            sage: H._get_matrix([8..23])
            [ 8  9 10 11]
            [12 13 14 15]
            [16 17 18 19]
            [20 21 22 23]
        """
        try:
            if g.parent() is self.matrix_space():
                return g
        except AttributeError:
            pass

        if isinstance(g, morphism.Morphism):
            return g.matrix()
        elif hasattr(g, 'list'):
            return self.matrix_space()(g.list())
        else:
            return self.matrix_space()(g)

    def free_module(self):
        r"""
        Return this endomorphism ring as a free submodule of a big
        `\ZZ^{4nm}`, where `n` is the dimension of
        the domain abelian variety and `m` the dimension of the
        codomain.

        OUTPUT: free module

        EXAMPLES::

            sage: E = Hom(J0(11), J0(22))
            sage: E.free_module()
            Free module of degree 8 and rank 2 over Integer Ring
            Echelon basis matrix:
            [ 1  0 -3  1  1  1 -1 -1]
            [ 0  1 -3  1  1  1 -1  0]
        """
        self.calculate_generators()
        V = ZZ**(4*self.domain().dimension() * self.codomain().dimension())
        return V.submodule([V(m.matrix().list()) for m in self.gens()])

    def gen(self, i=0):
        """
        Return `i`-th generator of ``self``.

        INPUT:

        - ``i`` -- integer

        OUTPUT: a morphism

        EXAMPLES::

            sage: E = End(J0(22))
            sage: E.gen(0).matrix()
            [1 0 0 0]
            [0 1 0 0]
            [0 0 1 0]
            [0 0 0 1]
        """
        self.calculate_generators()
        if i > self.ngens():
            raise ValueError("self only has %s generators" % self.ngens())
        return self.element_class(self, self._gens[i])

    def ngens(self):
        """
        Return number of generators of ``self``.

        OUTPUT: integer

        EXAMPLES::

            sage: E = End(J0(22))
            sage: E.ngens()
            4
        """
        self.calculate_generators()
        return len(self._gens)

    def gens(self) -> tuple:
        """
        Return tuple of generators for this endomorphism ring.

        EXAMPLES::

            sage: E = End(J0(22))
            sage: E.gens()
            (Abelian variety endomorphism of Abelian variety J0(22) of dimension 2,
             Abelian variety endomorphism of Abelian variety J0(22) of dimension 2,
             Abelian variety endomorphism of Abelian variety J0(22) of dimension 2,
             Abelian variety endomorphism of Abelian variety J0(22) of dimension 2)
        """
        try:
            return self._gen_morphisms
        except AttributeError:
            self.calculate_generators()
            self._gen_morphisms = tuple([self.gen(i) for i in range(self.ngens())])
            return self._gen_morphisms

    def matrix_space(self):
        """
        Return the underlying matrix space that we view this endomorphism
        ring as being embedded into.

        EXAMPLES::

            sage: E = End(J0(22))
            sage: E.matrix_space()
            Full MatrixSpace of 4 by 4 dense matrices over Integer Ring
        """
        return self._matrix_space

    def calculate_generators(self):
        """
        If generators haven't already been computed, calculate generators
        for this homspace. If they have been computed, do nothing.

        EXAMPLES::

            sage: E = End(J0(11))
            sage: E.calculate_generators()
        """

        if self._gens is not None:
            return

        if (self.domain() == self.codomain()) and (self.domain().dimension() == 1):
            self._gens = (identity_matrix(ZZ, 2),)
            return

        phi = self.domain()._isogeny_to_product_of_powers()
        psi = self.codomain()._isogeny_to_product_of_powers()

        H_simple = phi.codomain().Hom(psi.codomain())
        im_gens = H_simple._calculate_product_gens()

        M = phi.matrix()
        Mt = psi.complementary_isogeny().matrix()

        R = ZZ**(4*self.domain().dimension()*self.codomain().dimension())
        gens = R.submodule([(M*self._get_matrix(g)*Mt).list()
                            for g in im_gens]).saturation().basis()
        self._gens = tuple([self._get_matrix(g) for g in gens])

    def _calculate_product_gens(self):
        """
        For internal use.

        Calculate generators for ``self``, assuming that ``self`` is a product
        of simple factors.

        EXAMPLES::

            sage: E = End(J0(37))
            sage: E.gens()
            (Abelian variety endomorphism of Abelian variety J0(37) of dimension 2,
             Abelian variety endomorphism of Abelian variety J0(37) of dimension 2)
            sage: [ x.matrix() for x in E.gens() ]
            [
            [1 0 0 0]  [ 0  1  1 -1]
            [0 1 0 0]  [ 1  0  1  0]
            [0 0 1 0]  [ 0  0 -1  1]
            [0 0 0 1], [ 0  0  0  1]
            ]
            sage: E._calculate_product_gens()
            [
            [1 0 0 0]  [0 0 0 0]
            [0 1 0 0]  [0 0 0 0]
            [0 0 0 0]  [0 0 1 0]
            [0 0 0 0], [0 0 0 1]
            ]
        """
        Afactors = self.domain().decomposition(simple=False)
        Bfactors = self.codomain().decomposition(simple=False)
        if len(Afactors) == 1 and len(Bfactors) == 1:
            Asimples = Afactors[0].decomposition()
            Bsimples = Bfactors[0].decomposition()
            if len(Asimples) == 1 and len(Bsimples) == 1:
                # Handle the base case of A, B simple
                gens = self._calculate_simple_gens()

            else:
                # Handle the case of A, B simple powers
                gens = []
                phi_matrix = Afactors[0]._isogeny_to_product_of_simples().matrix()
                psi_t_matrix = Bfactors[0]._isogeny_to_product_of_simples().complementary_isogeny().matrix()
                for i in range(len(Asimples)):
                    for j in range(len(Bsimples)):
                        hom_gens = Asimples[i].Hom(Bsimples[j]).gens()
                        for sub_gen in hom_gens:
                            sub_mat = sub_gen.matrix()
                            M = copy(self.matrix_space().zero_matrix())
                            M.set_block(sub_mat.nrows()*i, sub_mat.ncols()*j, sub_mat)
                            gens.append(phi_matrix * M * psi_t_matrix)

        else:
            # Handle the case of A, B generic
            gens = []
            cur_row = 0
            for Afactor in Afactors:
                cur_row += Afactor.dimension() * 2
                cur_col = 0
                for Bfactor in Bfactors:
                    cur_col += Bfactor.dimension() * 2
                    Asimple = Afactor[0]
                    Bsimple = Bfactor[0]
                    if Asimple.newform_label() == Bsimple.newform_label():
                        for sub_gen in Afactor.Hom(Bfactor).gens():
                            sub_mat = sub_gen.matrix()
                            M = copy(self.matrix_space().zero_matrix())
                            M.set_block(cur_row - sub_mat.nrows(),
                                        cur_col - sub_mat.ncols(),
                                        sub_mat)
                            gens.append(M)

        return gens

    def _calculate_simple_gens(self):
        """
        Calculate generators for ``self``, where both the domain and codomain
        for ``self`` are assumed to be simple abelian varieties.

        The saturation of the span of these generators in ``self``
        will be the full space of homomorphisms from the domain of
        ``self`` to its codomain.

        EXAMPLES::

            sage: H = Hom(J0(11), J0(22)[0])
            sage: H._calculate_simple_gens()
            [
            [1 0]
            [1 1]
            ]
            sage: J = J0(11) * J0(33) ; J.decomposition()
            [Simple abelian subvariety 11a(1,11) of dimension 1 of J0(11) x J0(33),
             Simple abelian subvariety 11a(1,33) of dimension 1 of J0(11) x J0(33),
             Simple abelian subvariety 11a(3,33) of dimension 1 of J0(11) x J0(33),
             Simple abelian subvariety 33a(1,33) of dimension 1 of J0(11) x J0(33)]
            sage: J[0].Hom(J[1])._calculate_simple_gens()
            [
            [ 0 -1]
            [ 1 -1]
            ]
            sage: J[0].Hom(J[2])._calculate_simple_gens()
            [
            [-1  0]
            [-1 -1]
            ]
            sage: J[0].Hom(J[0])._calculate_simple_gens()
            [
            [1 0]
            [0 1]
            ]
            sage: J[1].Hom(J[2])._calculate_simple_gens()
            [
            [ 0 -4]
            [ 4  0]
            ]

        ::

            sage: J = J0(23) ; J.decomposition()
            [Simple abelian variety J0(23) of dimension 2]
            sage: J[0].Hom(J[0])._calculate_simple_gens()
            [
            [1 0 0 0]  [ 0  1 -1  0]
            [0 1 0 0]  [ 0  1 -1  1]
            [0 0 1 0]  [-1  2 -2  1]
            [0 0 0 1], [-1  1  0 -1]
            ]
            sage: J.hecke_operator(2).matrix()
            [ 0  1 -1  0]
            [ 0  1 -1  1]
            [-1  2 -2  1]
            [-1  1  0 -1]

        ::

            sage: H = Hom(J0(11), J0(22)[0])
            sage: H._calculate_simple_gens()
            [
            [1 0]
            [1 1]
            ]
        """
        A = self.domain()
        B = self.codomain()

        if A.newform_label() != B.newform_label():
            return []

        f = A._isogeny_to_newform_abelian_variety()
        g = B._isogeny_to_newform_abelian_variety().complementary_isogeny()

        Af = f.codomain()
        ls = Af._calculate_endomorphism_generators()

        Mf = f.matrix()
        Mg = g.matrix()

        return [Mf * self._get_matrix(e) * Mg for e in ls]


# NOTE/WARNING/TODO:  Below in the __init__, etc. we do *not* check
# that the input gens are give something that spans a sub*ring*, as opposed
# to just a subgroup.
class EndomorphismSubring(Homspace):

    def __init__(self, A, gens=None, category=None):
        """
        A subring of the endomorphism ring.

        INPUT:

        - ``A`` -- an abelian variety

        - ``gens`` -- (default: ``None``) if given
          should be a tuple of the generators as matrices

        EXAMPLES::

            sage: J0(23).endomorphism_ring()
            Endomorphism ring of Abelian variety J0(23) of dimension 2
            sage: sage.modular.abvar.homspace.EndomorphismSubring(J0(25))
            Endomorphism ring of Abelian variety J0(25) of dimension 0
            sage: E = J0(11).endomorphism_ring()
            sage: type(E)
            <class 'sage.modular.abvar.homspace.EndomorphismSubring_with_category'>
            sage: E.homset_category()
            Category of modular abelian varieties over Rational Field
            sage: E.category()
            Category of endsets of modular abelian varieties over Rational Field
            sage: E in Rings()
            True
            sage: TestSuite(E).run(skip=["_test_prod"])

        TESTS:

        The following tests against a problem on 32 bit machines that
        occurred while working on :issue:`9944`::

            sage: sage.modular.abvar.homspace.EndomorphismSubring(J1(12345))
            Endomorphism ring of Abelian variety J1(12345) of dimension 5405473

        :issue:`16275` removed the custom ``__reduce__`` method, since
        :meth:`Homset.__reduce__` already implements appropriate
        unpickling by construction::

            sage: E.__reduce__.__module__
            'sage.categories.homset'
            sage: E.__reduce__()
            (<function Hom at ...>,
             (Abelian variety J0(11) of dimension 1,
              Abelian variety J0(11) of dimension 1,
              Category of modular abelian varieties over Rational Field,
             False))
        """
        self._J = A.ambient_variety()
        self._A = A

        # Initialise self with the correct category.
        if category is None:
            homset_cat = A.category()
        else:
            homset_cat = category
        # Remark: Parent.__init__ will automatically form the join
        # of the category of rings and of homset_cat
        Parent.__init__(self, A.base_ring(), category=homset_cat.Endsets())
        Homspace.__init__(self, A, A, cat=homset_cat)
        if gens is None:
            self._gens = None
        else:
            self._gens = tuple([self._get_matrix(g) for g in gens])
        self._is_full_ring = gens is None

    def _repr_(self):
        """
        Return the string representation of ``self``.

        EXAMPLES::

            sage: J0(31).endomorphism_ring()._repr_()
            'Endomorphism ring of Abelian variety J0(31) of dimension 2'
            sage: J0(31).endomorphism_ring().image_of_hecke_algebra()._repr_()
            'Subring of endomorphism ring of Abelian variety J0(31) of dimension 2'
        """
        if self._is_full_ring:
            return "Endomorphism ring of %s" % self._A
        else:
            return "Subring of endomorphism ring of %s" % self._A

    def abelian_variety(self):
        """
        Return the abelian variety that this endomorphism ring is attached
        to.

        EXAMPLES::

            sage: J0(11).endomorphism_ring().abelian_variety()
            Abelian variety J0(11) of dimension 1
        """
        return self._A

    def index_in(self, other, check=True):
        """
        Return the index of ``self`` in ``other``.

        INPUT:

        - ``other`` -- another endomorphism subring of the
          same abelian variety

        - ``check`` -- boolean (default: ``True``); whether to do some
          type and other consistency checks

        EXAMPLES::

            sage: R = J0(33).endomorphism_ring()
            sage: R.index_in(R)
            1
            sage: J = J0(37) ; E = J.endomorphism_ring() ; T = E.image_of_hecke_algebra()
            sage: T.index_in(E)
            1
            sage: J = J0(22) ; E = J.endomorphism_ring() ; T = E.image_of_hecke_algebra()
            sage: T.index_in(E)
            +Infinity
        """
        if check:
            if not isinstance(other, EndomorphismSubring):
                raise ValueError("other must be a subring of an endomorphism ring of an abelian variety.")
            if not (self.abelian_variety() == other.abelian_variety()):
                raise ValueError("self and other must be endomorphisms of the same abelian variety")

        M = self.free_module()
        N = other.free_module()
        if M.rank() < N.rank():
            return Infinity
        return M.index_in(N)

    def index_in_saturation(self):
        """
        Given a Hecke algebra T, compute its index in its saturation.

        EXAMPLES::

            sage: End(J0(23)).image_of_hecke_algebra().index_in_saturation()
            1
            sage: End(J0(44)).image_of_hecke_algebra().index_in_saturation()
            2
        """
        A = self.abelian_variety()
        d = A.dimension()
        M = ZZ**(4*d**2)
        gens = [x.matrix().list() for x in self.gens()]
        R = M.submodule(gens)
        return R.index_in_saturation()

    def discriminant(self):
        """
        Return the discriminant of this ring, which is the discriminant of
        the trace pairing.

        .. NOTE::

           One knows that for modular abelian varieties, the
           endomorphism ring should be isomorphic to an order in a
           number field. However, the discriminant returned by this
           function will be `2^n` ( `n =`
           self.dimension()) times the discriminant of that order,
           since the elements are represented as 2d x 2d
           matrices. Notice, for example, that the case of a one
           dimensional abelian variety, whose endomorphism ring must
           be ZZ, has discriminant 2, as in the example below.

        EXAMPLES::

            sage: J0(33).endomorphism_ring().discriminant()
            -64800
            sage: J0(46).endomorphism_ring().discriminant()  # long time (6s on sage.math, 2011)
            24200000000
            sage: J0(11).endomorphism_ring().discriminant()
            2
        """
        g = self.gens()
        M = matrix(ZZ, len(g), [(g[i]*g[j]).trace()
                                for i in range(len(g)) for j in range(len(g))])
        return M.determinant()

    def image_of_hecke_algebra(self, check_every=1):
        """
        Compute the image of the Hecke algebra inside this endomorphism
        subring.

        We simply calculate Hecke operators up to the Sturm bound, and look
        at the submodule spanned by them. While computing, we can check to
        see if the submodule spanned so far is saturated and of maximal
        dimension, in which case we may be done. The optional argument
        check_every determines how many Hecke operators we add in before
        checking to see if this condition is met.

        INPUT:

        - ``check_every`` -- integer (default: 1); if this integer is positive,
          this integer determines how many Hecke operators we add in before
          checking to see if the submodule spanned so far is maximal and
          saturated

        OUTPUT: the image of the Hecke algebra as a subring of ``self``

        EXAMPLES::

            sage: E = J0(33).endomorphism_ring()
            sage: E.image_of_hecke_algebra()
            Subring of endomorphism ring of Abelian variety J0(33) of dimension 3
            sage: E.image_of_hecke_algebra().gens()
            (Abelian variety endomorphism of Abelian variety J0(33) of dimension 3,
             Abelian variety endomorphism of Abelian variety J0(33) of dimension 3,
             Abelian variety endomorphism of Abelian variety J0(33) of dimension 3)
            sage: [ x.matrix() for x in E.image_of_hecke_algebra().gens() ]
            [
            [1 0 0 0 0 0]  [ 0  2  0 -1  1 -1]  [ 0  0  1 -1  1 -1]
            [0 1 0 0 0 0]  [-1 -2  2 -1  2 -1]  [ 0 -1  1  0  1 -1]
            [0 0 1 0 0 0]  [ 0  0  1 -1  3 -1]  [ 0  0  1  0  2 -2]
            [0 0 0 1 0 0]  [-2  2  0  1  1 -1]  [-2  0  1  1  1 -1]
            [0 0 0 0 1 0]  [-1  1  0  2  0 -3]  [-1  0  1  1  0 -1]
            [0 0 0 0 0 1], [-1  1 -1  1  1 -2], [-1  0  0  1  0 -1]
            ]
            sage: J0(33).hecke_operator(2).matrix()
            [-1  0  1 -1  1 -1]
            [ 0 -2  1  0  1 -1]
            [ 0  0  0  0  2 -2]
            [-2  0  1  0  1 -1]
            [-1  0  1  1 -1 -1]
            [-1  0  0  1  0 -2]
        """
        try:
            return self.__hecke_algebra_image
        except AttributeError:
            pass

        A = self.abelian_variety()
        if not A.is_hecke_stable():
            raise ValueError("ambient variety is not Hecke stable")

        M = A.modular_symbols()

        d = A.dimension()
        EndVecZ = ZZ**(4*d**2)

        if d == 1:
            self.__hecke_algebra_image = EndomorphismSubring(A, [[1, 0, 0, 1]])
            return self.__hecke_algebra_image

        V = EndVecZ.submodule([A.hecke_operator(1).matrix().list()])

        for n in range(2, M.sturm_bound()+1):
            if (check_every > 0 and
                    n % check_every == 0 and
                    V.dimension() == d and
                    V.index_in_saturation() == 1):
                break
            V += EndVecZ.submodule([A.hecke_operator(n).matrix().list()])

        self.__hecke_algebra_image = EndomorphismSubring(A, V.basis())
        return self.__hecke_algebra_image



# sage_setup: distribution = sagemath-cmr
from sage.libs.cmr.cmr cimport CMR_SEYMOUR_NODE, CMR_ELEMENT
from sage.structure.sage_object cimport SageObject


cdef class DecompositionNode(SageObject):
    cdef object _base_ring
    cdef object _matrix
    cdef CMR_SEYMOUR_NODE *_dec
    cdef object _row_keys
    cdef object _column_keys
    cdef object _child_nodes
    cdef object _minors

    cdef _set_dec(self, CMR_SEYMOUR_NODE *dec)
    cdef _set_root_dec(self)
    cdef _set_row_keys(self, row_keys)
    cdef _set_column_keys(self, column_keys)

    cdef _CMRelement_to_key(self, CMR_ELEMENT element)


cdef class BaseGraphicNode(DecompositionNode):
    cdef object _graph
    cdef object _forest_edges
    cdef object _coforest_edges


cdef class GraphicNode(BaseGraphicNode):
    pass


cdef class CographicNode(BaseGraphicNode):
    pass


cdef class PlanarNode(BaseGraphicNode):
    cdef object _cograph
    cdef object _cograph_forest_edges
    cdef object _cograph_coforest_edges


cdef create_DecompositionNode(CMR_SEYMOUR_NODE *dec,
                              matrix=?,
                              row_keys=?, column_keys=?,
                              base_ring=?)
